from chameleon.core import template
from chameleon.zpt.language import Parser
from chameleon.zpt.language import TextParser

class PageTemplate(template.Template):
    __doc__ = template.Template.__doc__ # for Sphinx autodoc
    default_parser = Parser()
    version = 3

    def __init__(self, body, parser=None, **kwargs):
        kwargs['parser']=parser is not None and parser or self.default_parser
        super(PageTemplate, self).__init__(body, **kwargs)

    def render(self, *args, **kwargs):
        kwargs.setdefault("template", self)
        kwargs.setdefault("macros", self.macros)
        return super(PageTemplate, self).render(*args, **kwargs)

    def render_macro(self, macro, global_scope=False, slots={}, parameters={}):
        parameters.setdefault("macros", self.macros)
        return super(PageTemplate, self).render_macro(
            macro, global_scope=global_scope, slots=slots, parameters=parameters)

class PageTemplateFile(template.TemplateFile, PageTemplate):
    __doc__ = template.TemplateFile.__doc__ # for Sphinx autodoc

    default_parser = Parser()

    def __init__(self, filename, parser=None, **kwargs):
        kwargs['parser']=parser is not None and parser or self.default_parser
        super(PageTemplateFile, self).__init__(filename, **kwargs)

class PageTextTemplate(PageTemplate):
    __doc__ = template.Template.__doc__ # for Sphinx autodoc
    default_parser = TextParser()
    format = 'text'

class PageTextTemplateFile(PageTemplateFile):
    __doc__ = template.Template.__doc__ # for Sphinx autodoc
    default_parser = TextParser()
    format = 'text'




import re

from chameleon.core.utils import implements
from chameleon.core.utils import queryUtility
from chameleon.core.utils import queryAdapter
from chameleon.core.utils import adapts
from chameleon.core import types
from chameleon.core import parsing
from chameleon.ast.astutil import parse

import interfaces

translators = {
    'python': lambda translator: python_translator,
    'string': lambda translator: StringTranslator(translator),
    'import': lambda translator: ImportTranslator()}

def lookup_translator(current, name):
    """Look up translator. Priority is given to translators registered
    as components. If a component lookup fails, use the module-global
    ``translators`` dictionary.

    The recommended way to add new translators is to use the component
    architecture.
    """

    translator = queryUtility(
        interfaces.IExpressionTranslator, name=name) or \
        queryAdapter(
        current, interfaces.IExpressionTranslator, name=name)

    if translator is not None:
        return translator

    function = translators.get(name)
    if function is not None:
        return function(current)

class ExpressionTranslator(object):
    """Base class for TALES expression translation."""

    implements(interfaces.IExpressionTranslator)

    re_pragma = re.compile(r'^\s*(?P<pragma>[a-z]+):')
    re_method = re.compile(r'^(?P<name>[A-Za-z0-9_]+)'
                           '(\((?P<args>[A-Za-z0-9_]+\s*(,\s*[A-Za-z0-9_]+)*)\))?')

    recursive = False

    def __init__(self):
        self.translator = self

    def pragma(self, name):
        return lookup_translator(self, name)

    def declaration(self, string):
        """Variable declaration.

        >>> declaration = ExpressionTranslator().declaration

        Single variable:

        >>> declaration("variable")
        declaration('variable')

        Multiple variables:

        >>> declaration("variable1, variable2")
        declaration('variable1', 'variable2')

        >>> declaration("variable1,variable2")
        declaration('variable1', 'variable2')

        Repeat not allowed:

        >>> declaration('repeat')
        Traceback (most recent call last):
         ...
        ValueError: Invalid variable name 'repeat' (reserved).

        >>> declaration('_disallowed')
        Traceback (most recent call last):
         ...
        ValueError: Invalid variable name '_disallowed' (starts with an underscore).
        """

        variables = []
        for var in string.split(','):
            var = var.strip()

            if var in ('repeat',):
                raise ValueError, "Invalid variable name '%s' (reserved)." % var

            if var.startswith('_') and not var.startswith('_tmp'):
                raise ValueError(
                    "Invalid variable name '%s' (starts with an underscore)." % var)

            variables.append(var)

        return types.declaration(variables)

    def mapping(self, string):
        """Semicolon-separated mapping.
        
        >>> mapping = ExpressionTranslator().mapping

        >>> mapping("abc def")
        mapping(('abc', 'def'),)

        >>> mapping("abc def;")
        mapping(('abc', 'def'),)

        >>> mapping("abc")
        mapping(('abc', None),)

        >>> mapping("abc;")
        mapping(('abc', None),)

        >>> mapping("abc; def ghi")
        mapping(('abc', None), ('def', 'ghi'))
        """

        defs = string.split(';')
        mappings = []
        for d in defs:
            d = d.strip()
            if d == '':
                continue
            parts = d.split()
            if len(parts) == 1:
                mappings.append((d, None))
            elif len(parts) == 2:
                mappings.append((parts[0], parts[1]))
            else:
                raise ValueError, "Invalid mapping (%s)." % string

        return types.mapping(mappings)

    def definitions(self, string):
        """Semi-colon separated variable definitions.
        
        >>> class MockExpressionTranslator(ExpressionTranslator):
        ...     def tales(self, string, escape=None):
        ...         if string == '' or ';' in string.replace("';'", "''"):
        ...             raise SyntaxError()
        ...         return types.value(string.strip())

        >>> definitions = MockExpressionTranslator().definitions
        
        Single define:
        
        >>> definitions("variable expression")
        definitions((declaration('variable'), value('expression')),)
        
        Multiple defines:
        
        >>> definitions("variable1 expression1; variable2 expression2")
        definitions((declaration('variable1'), value('expression1')),
                    (declaration('variable2'), value('expression2')))

        Defines are only split on semi-colon if a valid declaration is
        available.
        
        >>> definitions("variable1 ';'+expression1; variable2 expression2")
        definitions((declaration('variable1'), value("';'+expression1")),
                    (declaration('variable2'), value('expression2')))

        Tuple define:
        
        >>> definitions("(variable1, variable2) (expression1, expression2)")
        definitions((declaration('variable1', 'variable2'),
                    value('(expression1, expression2)')),)

        Global defines:

        >>> definitions("global variable expression")
        definitions((declaration('variable', global_scope=True), value('expression')),)

        >>> definitions("variable1 expression1; global variable2 expression2")
        definitions((declaration('variable1'), value('expression1')),
                    (declaration('variable2', global_scope=True), value('expression2')))

        Space, the 'in' operator and '=' may be used to separate
        variable from expression.

        >>> definitions("variable in expression")
        definitions((declaration('variable'), value('expression')),)        
        
        >>> definitions("variable1 = expression1; variable2 = expression2")
        definitions((declaration('variable1'), value('expression1')),
                    (declaration('variable2'), value('expression2')))

        >>> definitions("variable1=expression1; variable2=expression2")
        definitions((declaration('variable1'), value('expression1')),
                    (declaration('variable2'), value('expression2')))
        
        A define clause that ends in a semicolon:
        
        >>> definitions("variable expression;")
        definitions((declaration('variable'), value('expression')),)
        
        A define clause with a trivial expression (we do allow this):
        
        >>> definitions("variable")
        definitions((declaration('variable'), None),)
        
        A proper define clause following one with a trivial expression:
        
        >>> definitions("variable1 expression; variable2")
        definitions((declaration('variable1'), value('expression')),
                    (declaration('variable2'), None))
        """

        string = string.replace('\n', '').strip()

        defines = []
        i = 0
        while i < len(string):
            while string[i] == ' ':
                i += 1

            global_scope = False
            if string[i:].startswith('global'):
                global_scope = True
                i += 6

            # get variable definition
            if string[i] == '(':
                j = string.find(')', i+1)
                if j == -1:
                    raise ValueError, "Invalid variable tuple definition (%s)." % string
                var = self.declaration(string[i+1:j])
                j += 1
            else:
                j = string.find('=', i + 1)
                k = string.find(' ', i + 1)
                if k < j and k > -1 or j < 0:
                    j = k
                
                if j < 0:
                    var = self.declaration(string[i:])
                    j = len(string)
                else:
                    var = self.declaration(string[i:j])

            var.global_scope = global_scope

            # get expression
            i = j + len(string) - j - len(string[j:].lstrip())

            token = string[i:]
            if token.startswith('=='):
                raise ValueError("Invalid variable definition (%s)." % string)
            elif token.startswith('='):
                i += 1
            elif token.startswith('in '):
                i += 3

            try:
                expr = self.tales(string[i:], ';')
                j = -1
            except SyntaxError:
                expr = None
                j = len(string)

            while j > i:
                j = string.rfind(';', i, j)
                if j < 0:
                    # this is always a re-raise from right above
                    raise

                try:
                    expr = self.tales(string[i:j], ';')
                except SyntaxError:
                    if string.rfind(';', i, j) > 0:
                        continue
                    raise

                break

            defines.append((var, expr))

            if j < 0:
                break

            i = j + 1

        return types.definitions(defines)

    def definition(self, string):
        defs = self.definitions(string)
        if len(defs) != 1:
            raise ValueError, "Multiple definitions not allowed."

        return defs[0]

    def output(self, string):
        """String output; supports 'structure' keyword.
        
        >>> class MockExpressionTranslator(ExpressionTranslator):
        ...     def translate(self, string, escape=None):
        ...         if string: return types.value(string)

        >>> output = MockExpressionTranslator().output

        >>> output("context/title")
        escape(value('context/title'),)

        >>> output("context/pretty_title_or_id|context/title")
        escape(value('context/pretty_title_or_id'), value('context/title'))

        >>> output("structure context/title")
        value('context/title')        

        >>> output(" structure context/title")
        value('context/title')

        """

        string = string.lstrip(' ')
        if string.startswith('structure '):
            return self.tales(string[len('structure'):])
        
        expression = self.tales(string)

        if isinstance(expression, types.parts):
            return types.escape(expression)

        return types.escape((expression,))
            
    def tales(self, string, escape=None):
        """We need to implement the ``validate`` and
        ``translate``-methods. Let's define that an expression is
        valid if it contains an odd number of characters.
        
        >>> class MockExpressionTranslator(ExpressionTranslator):
        ...     def translate(self, string, escape=None):
        ...         if string: return types.value(string)

        >>> tales = MockExpressionTranslator().tales
                
        >>> tales('a')
        value('a')

        >>> tales('a|b')
        parts(value('a'), value('b'))
        """

        string = string.replace('\n', '').strip()

        if not string:
            return types.parts()

        parts = []

        # default translator is ``self``
        translator = self

        i = j = 0
        while i < len(string):
            match = self.re_pragma.match(string[i:])
            if match is not None:
                pragma = match.group('pragma')
                new_translator = self.pragma(pragma)
                if new_translator is not None:
                    translator = new_translator
                    i += match.end()
                    if translator.recursive:
                        value = translator.tales(string[i:])
                        if parts:
                            if isinstance(value, types.parts):
                                parts.extend(value)
                            else:
                                parts.append(value)

                            return types.parts(parts)
                        return value
                    continue

            j = string.find('|', j + 1)
            if j == -1:
                j = len(string)

            expr = string[i:j]

            try:
                value = translator.translate(expr, escape)
            except SyntaxError:
                if j < len(string):
                    continue
                raise

            value.label = expr
            parts.append(value)
            translator = self
            
            i = j + 1

        value = translator.translate("", escape)
        if value is not None:
            value.label = ""
            parts.append(value)

        if len(parts) == 1:
            return parts[0]

        return types.parts(parts)

    def split(self, string):
        return parsing.interpolate(string, self.translator.tales)

class PythonTranslator(ExpressionTranslator):
    """Implements Python expression translation."""

    def translate(self, string, escape=None):
        """We use the ``parser`` module to determine if
        an expression is a valid python expression.

        Make sure the syntax error exception contains the expression
        string.

        >>> translate = PythonTranslator().translate
        >>> try: translate('abc:def:ghi')
        ... except SyntaxError, e: 'abc:def:ghi' in repr(e)
        True
        """

        if isinstance(string, unicode):
            string = string.encode('utf-8')

        if string:
            expression = string.strip()
            parse(expression, 'eval')

            if isinstance(string, str):
                string = string.decode('utf-8')

            return types.value(string.strip())

python_translator = PythonTranslator()

class StringTranslator(ExpressionTranslator):
    """Implements string translation expression."""

    adapts(interfaces.IExpressionTranslator)
    
    re_interpolation = re.compile(r'(?P<prefix>[^\\]\$|^\$)({((?P<expression>.*)})?|'
                                  '(?P<variable>[A-Za-z][A-Za-z0-9_]*))')

    def __init__(self, translator):
        self.translator = translator

    def translate(self, string, escape=None):
        """
        >>> translate = StringTranslator(python_translator).translate

        >>> translate("")
        join('',)
        
        """

        if escape is not None:
            string = string.rstrip(escape)
            
        parts = self.split(string)
        if escape is not None:
            parts = map(
                lambda part: isinstance(part, types.expression) and \
                part or self._unescape(part, escape), parts)

        return types.join(parts)            

    def _unescape(self, string, symbol):
        """
        >>> unescape = StringTranslator(None)._unescape
        
        >>> unescape('string:Hello World', ';')
        'string:Hello World'

        >>> unescape('string:Hello World;', ';')
        'string:Hello World;'

        >>> unescape('string:Hello World;;', ';')
        'string:Hello World;'
        
        >>> unescape('; string:Hello World', ';')
        Traceback (most recent call last):
         ...
        SyntaxError: Must escape symbol ';'.

        >>> unescape('; string:Hello World;', ';')
        Traceback (most recent call last):
         ...
        SyntaxError: Must escape symbol ';'.

        >>> unescape('string:Hello; World', ';')
        Traceback (most recent call last):
         ...
        SyntaxError: Must escape symbol ';'.

        >>> unescape(';; string:Hello World', ';')
        '; string:Hello World'
        """

        if re.search("([^%s]|^)%s[^%s]" % (symbol, symbol, symbol), string):
            raise SyntaxError(
                "Must escape symbol %s." % repr(symbol))
        
        return string.replace(symbol+symbol, symbol)

class ImportTranslator(ExpressionTranslator):
    adapts(interfaces.IExpressionTranslator)
    symbol = '_resolve_dotted'
    re_dotted = re.compile(r'^[A-Za-z.]+$')

    def translate(self, string, escape=None):
        """
        >>> import_translator = ImportTranslator()
        >>> resolve_dotted = import_translator.translate

        >>> resolve_dotted("")

        >>> resolve_dotted("chameleon.zpt")
        value("_resolve_dotted('chameleon.zpt')")

        >>> translator = StringTranslator(import_translator)
        >>> translator.tales('import: chameleon.zpt')
        parts(value("_resolve_dotted('chameleon.zpt')"), join('',))

        >>> translator.definitions('zpt import: chameleon.zpt; core import: chameleon.core')
        definitions((declaration('zpt'), parts(value("_resolve_dotted('chameleon.zpt')"), join('',))), (declaration('core'), parts(value("_resolve_dotted('chameleon.core')"), join('',))))

        """

        if not string:
            return None

        string = string.strip()

        if self.re_dotted.match(string) is None:
            raise SyntaxError(string)

        value = types.value("%s('%s')" % (self.symbol, string))
        value.symbol_mapping[self.symbol] = _resolve_dotted
        return value

_module_cache = {}

# stolen from zope.dottedname
def resolve_dotted(name, module=None):
    name = name.split('.')
    if not name[0]:
        if module is None:
            raise ValueError("relative name without base module")
        module = module.split('.')
        name.pop(0)
        while not name[0]:
            module.pop()
            name.pop(0)
        name = module + name

    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used += '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)

    return found

def _resolve_dotted(dotted):
    if not dotted in _module_cache:
        resolved = resolve_dotted(dotted)
        _module_cache[dotted] = resolved
    return _module_cache[dotted]





from StringIO import StringIO
from cPickle import dumps

try:
    from hashlib import sha1 as sha
except ImportError:
    from sha import sha

try:
    import codegen
except ImportError:
    codegen = None

import generation
import clauses
import itertools
import types
import utils
import etree
import config
import copy
import i18n

GLOBALS = globals()

class Node(object):
    """Element translation class.

    This class implements the translation node for an element in the
    template document tree.

    It's used internally by the translation machinery.
    """

    symbols = config.SYMBOLS

    translate = None
    translation_name = None
    translation_domain = None
    translated_attributes = None
    skip = None
    cdata = None
    omit = None
    assign = None
    define = None
    macro = None
    use_macro = None
    extend_macro = None
    define_macro = None
    fill_slot = None
    define_slot = None
    condition = None
    repeat = None
    content = None
    include = None
    format = None
    dict_attributes = None
    dynamic_attributes = utils.emptydict

    ns_omit = "http://xml.zope.org/namespaces/meta",

    def __init__(self, element):
        self.element = element

    @property
    def tag(self):
        tag = self.element.tag
        if '}' in tag:
            url, tag = tag[1:].split('}')
            for prefix, _url in self.element.nsmap.items():
                if prefix is not None and _url == url:
                    return "%s:%s" % (prefix, tag)
        return tag

    @property
    def stream(self):
        return self.element.stream

    @property
    def text(self):
        if self.element.text is None:
            return ()
        if self.element.meta_structure is True:
            return (self.element.text,)
        else:
            return (utils.htmlescape(self.element.text),)

    @property
    def tail(self):
        if self.element.tail is None:
            return ()
        if self.element.meta_structure is True:
            return (self.element.tail,)
        else:
            return (utils.htmlescape(self.element.tail),)

    @property
    def ns_attributes(self):
        prefix_omit = set()
        ns_omit = list(self.ns_omit)

        root = self.element.getroottree().getroot()
        if root is None or utils.get_namespace(root) == config.META_NS:
            ns_omit.append("http://www.w3.org/1999/xhtml")

        prefix_omit = set()
        namespaces = self.element.nsmap.values()

        parent = self.element.getparent()
        while parent is not None:
            for prefix, ns in parent.nsmap.items():
                if ns in namespaces:
                    prefix_omit.add(prefix)
            parent = parent.getparent()

        attrs = dict(
            ((prefix and "xmlns:%s" % prefix or "xmlns", ns)
             for (prefix, ns) in self.element.nsmap.items()
             if ns not in ns_omit and prefix not in prefix_omit))

        return attrs

    @property
    def static_attributes(self):
        result = {}
        element_ns = self.element.nsmap.get(self.element.prefix)

        nsmap = self.element.nsmap.copy()
        nsmap.update(config.DEFAULT_NS_MAP)

        root = self.element.getroottree().getroot()
        omit_default_prefix = root.meta_omit_default_prefix

        for prefix, ns in nsmap.items():
            if ns not in self.ns_omit:
                attrs = utils.get_attributes_from_namespace(self.element, ns)
                if prefix is None or ns == element_ns:
                    attrs.update(
                        utils.get_attributes_from_namespace(self.element, None))

                for tag, value in attrs.items():
                    name = tag.split('}')[-1]

                    if prefix and (ns != element_ns or not omit_default_prefix):
                        result["%s:%s" % (prefix, name)] = value
                    elif omit_default_prefix:
                        result[name] = value

        if self.element.prefix is None:
            result.update(
                utils.get_attributes_from_namespace(self.element, None))

        return result

    def interpolated_attributes(self, internal):
        for name, value in self.element.attrib.items():
            if name in internal:
                continue
            if '}' in name:
                namespace, name = name[1:].split('}')
                if namespace not in self.ns_omit:
                    for prefix, ns in self.element.nsmap.items():
                        if ns == namespace:
                            break
                    else:
                        continue
                    name = "%s:%s" % (prefix, name)
            parts = self.element.translator.split(value)
            for part in parts:
                if isinstance(part, types.expression):
                    yield types.declaration((name,)), types.join(parts)
                    break

    @property
    def attribute_ordering(self):
        static = self.static_attributes
        ns = self.ns_attributes

        names = list(ns)

        for key in self.element.attrib:
            name = utils.format_attribute(self.element, key)

            if name in static:
                names.append(name)

        # ensure that an "xmlns" attribute comes first
        names.sort(key=lambda name: name == 'xmlns', reverse=True)

        return names

    def update(self):
        self.element.update()

    def begin(self):
        self.stream.scope.append(set())
        self.stream.begin(self.serialize())

    def end(self):
        self.stream.end(self.serialize())
        self.stream.scope.pop()

    def body(self):
        if isinstance(self.skip, types.expression):
            assert isinstance(self.skip, types.value), \
                   "Dynamic skip condition can't be of type %s." % type(self.skip)
            condition = clauses.Condition(self.skip, invert=True)
            if len(self.element):
                condition.begin(self.stream)
        elif self.skip:
            return

        for element in self.element:
            element.node.update()

        for element in self.element:
            element.node.visit()

        if isinstance(self.skip, types.expression):
            if len(self.element):
                condition.end(self.stream)

    def visit(self):
        assert self.stream is not None, "Must use ``start`` method."

        for element in self.element:
            if not isinstance(element, Element):
                self.wrap_literal(element)

        self.update()
        self.begin()
        self.body()
        self.end()

    def serialize(self):
        """Serialize element into clause-statements."""

        _ = []

        # i18n domain
        if self.translation_domain is not None:
            _.append(clauses.Define(
                self.symbols.domain, types.value(repr(self.translation_domain))))

        # variable definitions
        adding = set()
        if self.define is not None:
            # as an optimization, we only define the `default`
            # symbol, if it's present in the definition clause (as
            # string representation)
            if self.symbols.default in repr(self.define):
                default = types.value(self.symbols.default_marker_symbol)
                _.append(clauses.Assign(default, self.symbols.default))

            for declaration, expression in self.define:
                if self.symbols.remote_scope in self.stream.scope[1]:
                    define = clauses.Define(
                        declaration, expression, self.symbols.remote_scope)
                else:
                    define = clauses.Define(declaration, expression)

                for name in define.declaration:
                    adding.add(name)

                _.append(define)

        # tag tail (deferred)
        tail = self.tail
        if self.fill_slot is None and self.translation_name is None:
            for part in reversed(tail):
                if isinstance(part, types.expression):
                    _.append(clauses.Write(part, defer=True))
                else:
                    _.append(clauses.Out(part, defer=True))

        # macro method
        macro = self.macro
        if macro is not None:
            if self.symbols.remote_scope in self.stream.scope[1]:
                dictionary = self.symbols.remote_scope
            else:
                dictionary = self.symbols.scope

            exclude = set((
                self.symbols.scope, self.symbols.slots)) | \
                self.stream.scope[0] | set(macro.args)

            scope = set(itertools.chain(*self.stream.scope[1:])) | set((
                self.symbols.out, self.symbols.write))

            args = tuple(macro.args) + tuple(
                "%s=%s" % (name, name) for name in scope if name not in exclude)

            _.append(clauses.Method(
                macro.name, args,
                decorators=macro.decorators,
                dictionary=dictionary))

        # condition
        if self.condition is not None:
            _.append(clauses.Condition(self.condition))

        # repeat
        if self.repeat is not None:
            variables, expression = self.repeat
            newline = True
            for element in self.element.walk():
                node = element.node
                if node and node.omit is False:
                    break
            else:
                newline = False

            if len(variables) > 1:
                repeat = clauses.Repeat(
                    variables, expression, repeatdict=False, newline=newline)
            else:
                repeat = clauses.Repeat(variables, expression, newline=newline)
            _.append(repeat)

        # assign
        if self.assign is not None:
            for declaration, expression in self.assign:
                if len(declaration) != 1:
                    raise ValueError("Can only assign single variable.")
                variable = declaration[0]
                _.append(clauses.Assign(expression, variable))

        content = self.content
        omit = self.omit

        if self.define_slot:
            name = self.define_slot
            scope = set(itertools.chain(*self.stream.scope[1:]))
            # assemble arguments that we pass into the macro
            # fill-slot callback
            exclude = set((self.symbols.scope,
                           self.symbols.slots,
                           self.symbols.out,
                           self.symbols.write)).union(self.stream.scope[0])

            scope_args = tuple(variable for variable in scope
                               if variable not in exclude)

            # look up fill-slot value
            _.append(clauses.Assign(
                types.template('%%(slots)s.get(%s)' % repr(name)),
                self.symbols.tmp))

            # if slot has been filled, either use it as is (if
            # it's a string), or pass in required arguments (if
            # it's a callback function)
            _.append(clauses.Condition(
                types.template('%(tmp)s is not None'),
                (clauses.Condition(
                    types.template('isinstance(%(tmp)s, basestring)'),
                    (clauses.Slot(
                        types.template("%(tmp)s(%(scope)s)"),
                        scope_args),),
                    finalize=True,
                    invert=True),
                 clauses.Else((
                     clauses.Write(types.template("%(tmp)s")),))
                 )))

            _.append(clauses.Else())

        # set dynamic content flag
        dynamic = content or self.translate is not None

        # if an attribute ordering is required, setting a default
        # trivial value for each attribute will ensure that the order
        # is preserved
        attributes = utils.odict()
        if self.attribute_ordering is not None:
            for name in self.attribute_ordering:
                attributes[name] = None

        # static attributes (including those with a namespace prefix)
        # are at the bottom of the food chain
        attributes.update(self.static_attributes)
        attributes.update(self.ns_attributes)

        # dynamic attributes
        dynamic_attrs = self.dynamic_attributes or ()
        dynamic_attr_names = []

        for variables, expression in dynamic_attrs:
            if len(variables) != 1:
                raise ValueError("Tuple definitions in assignment clause "
                                     "is not supported.")

            variable = variables[0]
            attributes[variable] = expression
            dynamic_attr_names.append(variable)

        # translated attributes
        translated_attributes = self.translated_attributes or ()
        for variable, msgid in translated_attributes:
            if msgid:
                if variable in dynamic_attr_names:
                    raise ValueError(
                        "Message id not allowed in conjunction with "
                        "a dynamic attribute.")

                value = types.value('"%s"' % msgid)

                if variable in attributes:
                    default = repr(attributes[variable])
                    expression = self.translate_expression(value, default=default)
                else:
                    expression = self.translate_expression(value)
            else:
                value = attributes.get(variable)
                if value is not None:
                    if variable not in dynamic_attr_names:
                        value = repr(value)
                    expression = self.translate_expression(value)
                else:
                    raise ValueError("Must be either static or dynamic "
                                     "attribute when no message id "
                                     "is supplied.")

            attributes[variable] = expression

        # tag
        text = self.text
        if omit is not True:
            selfclosing = not text and not dynamic and len(self.element) == 0
            tag = clauses.Tag(
                self.tag, attributes,
                expression=self.dict_attributes, selfclosing=selfclosing,
                cdata=self.cdata is not None, defaults=self.static_attributes)
            if omit:
                _.append(clauses.Condition(
                    omit, [tag], finalize=False, invert=True))
            else:
                _.append(tag)

        # tag text (if we're not replacing tag body)
        if len(text) and not dynamic and not self.use_macro and not self.extend_macro:
            for part in text:
                if isinstance(part, types.expression):
                    _.append(clauses.Write(part))
                else:
                    _.append(clauses.Out(part))

        # dynamic content
        if content:
            msgid = self.translate
            if msgid is not None:
                if msgid:
                    raise ValueError(
                        "Can't use message id with dynamic content translation.")

                _.append(clauses.Assign(content, self.symbols.tmp))
                content = self.translate_expression(
                    types.value(self.symbols.tmp))
            else:
                value = types.value(repr(utils.serialize(self.element, omit=True)))
                _.insert(0, clauses.Assign(
                    value, "%s.value = %s" % (
                        self.symbols.default_marker_symbol, self.symbols.default)))

            _.append(clauses.Write(content))

        # dynamic text
        elif self.translate is not None and \
                 True in map(lambda part: isinstance(part, types.expression), text):
            if len(self.element):
                raise ValueError(
                    "Can't translate dynamic text block with elements in it.")

            init_stream = types.value('_init_stream()')
            init_stream.symbol_mapping['_init_stream'] = generation.initialize_stream

            subclauses = []
            subclauses.append(clauses.Define(
                types.declaration((self.symbols.out, self.symbols.write)), init_stream))

            for part in text:
                if isinstance(part, types.expression):
                    subclauses.append(clauses.Write(part))
                else:
                    part = ' '.join(part.split())
                    if part != "":
                        subclauses.append(clauses.Out(part))

            # attempt translation
            subclauses.append(clauses.Assign(
                self.translate_expression(
                types.value('%(out)s.getvalue()'),
                default=None), self.symbols.tmp))

            _.append(clauses.Group(subclauses))
            _.append(clauses.Write(types.value(self.symbols.tmp)))

        # include
        elif self.include:
            # compute macro function arguments and create argument string
            arguments = [
                "%s=%s" % (arg, arg) for arg in \
                set(itertools.chain(*self.stream.scope[1:]))]

            # XInclude's are similar to METAL macros, except the macro
            # is always defined as the entire template.

            # first we compute the filename expression and write it to
            # an internal variable
            _.append(clauses.Assign(self.include, self.symbols.include))

            # call template
            _.append(clauses.Write(
                types.template(
                "%%(xincludes)s.get(%%(include)s, %s).render_xinclude(%s)" % \
                (repr(self.format), ", ".join(arguments)))))

        # use or extend macro
        elif self.use_macro or self.extend_macro:
            # assign macro value to variable
            macro = self.use_macro or self.extend_macro
            _.append(clauses.Assign(macro, self.symbols.metal))

            # for each fill-slot element, create a new output stream
            # and save value in a temporary variable
            kwargs = []
            callbacks = {}

            # determine variable scope
            scope = set(itertools.chain(adding, *self.stream.scope[1:])) - \
                    self.stream.scope[0]

            # we pass in all variables from the current scope (as
            # keyword arguments, to allow first use before potential
            # reassignment)
            callback_args = ", ".join(
                "%s=%s" % (arg, arg) for arg in scope
                if arg not in (self.symbols.slots, self.symbols.scope))

            macro_args = ", ".join(
                "%s=%s" % (arg, arg) for arg in scope
                if arg not in (self.symbols.slots,))

            # loop through macro fill slot elements and generate
            # callback methods; the reason why we use callbacks is
            # convenience: it's an easy fit with the compiler
            elements = [element for element in self.element.walk()
                        if element.node and element.node.fill_slot]

            for element in elements:
                # make sure we're not in a nested macro block
                parent = element.getparent()
                while parent is not self.element:
                    if parent.node.use_macro or parent.node.extend_macro:
                        element = None
                        break
                    parent = parent.getparent()

                if element is None:
                    continue

                # determine and register callback name
                name = element.node.fill_slot
                callbacks[name] = callback = "%s_%s" % (
                    self.symbols.callback, utils.normalize_slot_name(name))

                # pass in remote scope to callback method; this is
                # done because macros may add global variables to the
                # scope, which should be made available to the calling
                # template
                visitor = clauses.Visit(element.node)
                tail = element.tail
                newline = tail and '\n' in tail
                _.append(clauses.Callback(
                    callback, visitor, callback_args, newline))

            # if we're extending the macro, the current slots will be
            # carried over to the macro
            extend = self.extend_macro is not None
            defines = set()

            if extend:
                for element in self.element.walk():
                    if element.node is not None:
                        define_slot = element.node.define_slot
                        if define_slot is not None:
                            defines.add(define_slot)

            # format slot arguments
            slot_args = ", ".join("'%s': %s" % kwarg for kwarg in callbacks.items())

            _.append(clauses.Macro(
                types.value("{%s}" % slot_args),
                macro_args,
                extend=extend,
                extend_except=defines,
                label=macro.label
                ))

        # translate body
        elif self.translate is not None:
            msgid = self.translate

            # subelements are either named or unnamed; if there are
            # unnamed elements, the message id must be dynamic
            named_elements = [e for e in self.element
                              if e.node.translation_name]

            unnamed_elements = [e for e in self.element
                                if not e.node.translation_name]

            if unnamed_elements and msgid:
                elements = ()
            else:
                elements = tuple(self.element)

            # if no message id is provided, there are two cases:
            #
            # 1) all dynamic content is assigned a translation name
            # 2) the message id needs to be generated dynamically
            #
            if not msgid and not unnamed_elements:
                msgid = self.create_msgid()

            if named_elements:
                mapping = self.symbols.mapping
                _.append(clauses.Assign(types.value('{}'), mapping))
            else:
                mapping = 'None'

            if not msgid:
                text = utils.htmlescape(self.element.text.replace('%', '%%') or "")
                _.append(clauses.Assign(types.value(repr(text)),
                                        self.symbols.msgid))

            # for each named block, create a new output stream
            # and use the value in the translation mapping dict
            for element in elements:
                init_stream = types.value('_init_stream()')
                init_stream.symbol_mapping[
                    '_init_stream'] = generation.initialize_stream

                subclauses = []
                subclauses.append(clauses.Define(
                    types.declaration((self.symbols.out, self.symbols.write)), init_stream))
                subclauses.append(clauses.Visit(element.node))

                # if the element is named, record it in the mapping
                if element in named_elements:
                    name = element.node.translation_name

                    subclauses.append(clauses.Assign(
                        types.template('%(out)s.getvalue()'),
                        "%s['%s']" % (mapping, name)))

                    # when computing a dynamic message id, add a
                    # reference to the named block
                    if not msgid:
                        if not unnamed_elements:
                            subclauses.append(clauses.Assign(
                                types.value(repr("${%s}" % name)), self.symbols.msgid))
                        else:
                            subclauses.append(clauses.Assign(
                                types.template('%(msgid)s + ' + repr("${%s}" % name) + ' + ' + repr(element.tail)),
                                self.symbols.msgid))

                # else add it to the dynamic message id
                else:
                    subclauses.append(clauses.Assign(
                        types.template('%(msgid)s + %(out)s.getvalue()'),
                        self.symbols.msgid))

                # XXX: note that this should read:
                # _.append(clauses.Group(subclauses))
                #
                # but there's a problem with multiple temporary
                # variable assignments within the same block; this is
                # just an easy work-around
                _.append(clauses.Condition(
                    types.value('True'), subclauses, finalize=True))

            if msgid:
                value = types.value(repr(msgid)).replace('%', '%%')
                if elements:
                    default = self.symbols.marker
                else:
                    default = types.value(repr(
                        utils.serialize(self.element, omit=True))).replace('%', '%%')
            else:
                default = types.template('%(msgid)s')
                value = types.template("' '.join(%(msgid)s.split())")

            _.append(clauses.Assign(
                self.translate_expression(
                    value, mapping=mapping, default=default), self.symbols.result))

            # write translation to output if successful, otherwise
            # fallback to default rendition; 
            result = types.value(self.symbols.result)
            result.symbol_mapping[self.symbols.marker] = i18n.marker

            if msgid and elements:
                condition = types.template('%(result)s is not %(marker)s')
                _.append(clauses.Condition(
                    condition, [clauses.UnicodeWrite(result)], finalize=True))

                subclauses = []
                if self.element.text:
                    subclauses.append(clauses.Out(
                        utils.htmlescape(self.element.text)))
                for element in self.element:
                    name = element.node.translation_name
                    if name:
                        value = types.value("%s['%s']" % (mapping, name))
                        subclauses.append(clauses.Write(value))
                    else:
                        subclauses.append(clauses.Out(utils.serialize(element)))

                    for part in reversed(element.node.tail):
                        if isinstance(part, types.expression):
                            subclauses.append(clauses.Write(part))
                        else:
                            subclauses.append(clauses.Out(
                                utils.htmlescape(part)))

                if subclauses:
                    _.append(clauses.Else(subclauses))
            else:
                _.append(clauses.UnicodeWrite(result))

        return _

    def wrap_literal(self, element):
        index = self.element.index(element)

        t = self.element.makeelement(utils.meta_attr('literal'))
        t.meta_omit = ""
        t.tail = element.tail
        t.text = unicode(element)
        for child in element.getchildren():
            t.append(child)
        self.element.remove(element)
        self.element.insert(index, t)
        t.update()

    def create_msgid(self):
        """Create an i18n msgid from the tag contents."""

        out = StringIO()
        out.write(self.element.text)

        for element in self.element:
            name = element.node.translation_name
            assert name is not None

            out.write("${%s}" % name)
            out.write(element.tail)

        msgid = out.getvalue().strip()
        msgid = ' '.join(msgid.split())

        return msgid

    def translate_expression(self, value, mapping=None, default=None):
        format = "%%(translate)s(%s, domain=%%(domain)s, mapping=%s, " \
                 "target_language=%%(language)s, default=%s)"
        return types.template(format % (value, mapping, default))

class Element(etree.ElementBase):
    """Template element class.

    To start translation at this element, use the ``start`` method,
    providing a code stream object.
    """

    translator = None

    class node(Node):
        @property
        def omit(self):
            if self.element.meta_omit is not None:
                return self.element.meta_omit or True
            if self.element.meta_replace:
                return True

        @property
        def content(self):
            return self.element.meta_replace

    node = property(node)

    def start(self, stream):
        self._stream = stream
        self.node.visit()

    def update(self):
        pass

    @property
    def root(self):
        tree = self.getroottree()
        root = tree.getroot()

        while root is not None:
            parent = root.getparent()
            if parent is None:
                break
            root = parent

        return root

    @property
    def stream(self):
        try:
            return self.root._stream
        except AttributeError:
            raise AttributeError("Can't locate stream object.")

    meta_cdata = etree.Annotation(
        utils.meta_attr('cdata'))

    meta_omit = etree.Annotation(
        utils.meta_attr('omit-tag'))

    meta_omit_default_prefix = etree.Annotation(
        utils.meta_attr('omit-default-prefix'))

    meta_structure = etree.Annotation(
        utils.meta_attr('structure'))

    meta_attributes = etree.Annotation(
        utils.meta_attr('attributes'))

    meta_replace = etree.Annotation(
        utils.meta_attr('replace'))

class MetaElement(Element):
    meta_cdata = etree.Annotation('cdata')

    meta_omit = True

    meta_structure = True

    meta_attributes = etree.Annotation('attributes')

    meta_replace = etree.Annotation('replace')

    meta_omit_default_prefix = etree.Annotation('omit-default-prefix')

class Compiler(object):
    """Template compiler.

    Initializes a template compiler for a template expressed as an
    ElementTree-document (``tree``).

    Document type may be configured using ``implicit_doctype`` will be
    used as the document type if the template does not define one
    itself, while ``explicit_doctype`` may be used to explicitly set a
    doctype regardless of what the template defines.

    The ``encoding`` parameter may be provided to return a document in
    a non-unicode encoding.

    By default, native attributes are printed with no prefix. This
    behavior may be overriden using the ``omit_default_prefix``
    parameter (certain legacy applications such as OpenOffice require
    this formatting).
    """

    doctype = None

    def __init__(self, tree, explicit_doctype=None,
                 xml_declaration=None, encoding=None, omit_default_prefix=True):
        self.xml_declaration = xml_declaration
        self.omit_default_prefix = omit_default_prefix
        self.tree = tree

        # explicit document type has priority over a parsed doctype
        if explicit_doctype is not None:
            self.doctype = explicit_doctype
        elif tree.docinfo.doctype:
            self.doctype = tree.docinfo.doctype

        if utils.coerces_gracefully(encoding):
            self.encoding = None
        else:
            self.encoding = encoding

        macros = self.macros = {}

        for element in tree.getroot().walk():
            node = element.node
            if node is not None and node.define_macro is not None:
                define = element.node.define_macro
                extend = element.node.extend_macro

                # collect slot names
                names = set()

                for subelement in element.walk():
                    node = subelement.node
                    if node is not None and node.define_slot is not None:
                        names.add(node.define_slot)

                macros[define] = (extend is not None, names)

    def __call__(self, macro, global_scope=True):
        root = copy.deepcopy(self.tree).getroot()

        if not isinstance(root, Element):
            raise ValueError(
                "Must define valid namespace for tag: '%s.'" % root.tag)

        # if macro is non-trivial, start compilation at the element
        # where the macro is defined
        if macro:
            for element in root.walk():
                node = element.node
                if node is not None and node.define_macro == macro:
                    element.meta_translator = root.meta_translator

                    # if element is the document root, render as a normal
                    # template, e.g. unset the `macro` mode
                    if root is element:
                        macro = None
                    else:
                        root = element

                    break
            else:
                raise KeyError(macro)

        # initialize code stream object
        stream = generation.CodeIO(
            root.node.symbols, encoding=self.encoding,
            indentation=0, indentation_string="\t")

        # transient symbols are added to the primary scope to exclude
        # them from being carried over when using macros
        for name, value in stream.symbols.as_dict().items():
            if value is config.TRANSIENT_SYMBOL:
                stream.scope[0].add(name)

        # initialize variable scope
        stream.scope.append(set(
            (stream.symbols.out,
             stream.symbols.write,
             stream.symbols.scope,
             stream.symbols.domain,
             stream.symbols.language)))

        if global_scope is False:
            stream.scope[-1].add(stream.symbols.remote_scope)

        # set up initialization code
        stream.symbol_mapping['_init_stream'] = generation.initialize_stream
        stream.symbol_mapping['_init_scope'] = generation.initialize_scope
        stream.symbol_mapping['_init_tal'] = generation.initialize_tal
        stream.symbol_mapping['_init_default'] = generation.initialize_default

        # add code-generation lookup globals
        stream.symbol_mapping.update(codegen.lookup_globals)

        if global_scope:
            assignments = (
                clauses.Assign(
                    types.value("_init_stream()"), ("%(out)s", "%(write)s")),
                clauses.Assign(
                     types.value("_init_tal()"), ("%(attributes)s", "%(repeat)s")),
                clauses.Assign(
                    types.value("_init_default()"), '%(default_marker_symbol)s'),
                clauses.Assign(
                    types.value(repr(None)), '%(default)s'),
                clauses.Assign(
                     types.template("None"), "%(domain)s"))
        else:
            assignments = (
                clauses.Assign(
                    types.template(
                        "%(scope)s['%(out)s'], %(scope)s['%(write)s']"),
                    ("%(out)s", "%(write)s")),
                clauses.Assign(
                    types.value("_init_tal()"), ("%(attributes)s", "%(repeat)s")),
                clauses.Assign(
                    types.value("_init_default()"), '%(default_marker_symbol)s'),
                clauses.Assign(
                    types.value(repr(None)), '%(default)s'),
                clauses.Assign(
                     types.template("None"), "%(domain)s"))

        for clause in assignments:
            clause.begin(stream)
            clause.end(stream)

        if macro is not None:
            nsmap = {}
            namespaces = set()

            for tag in root.attrib:
                if '}' not in tag:
                    continue

                namespace = tag[1:].split('}')[0]
                namespaces.add(namespace)

            for prefix, namespace in root.nsmap.items():
                if namespace in namespaces:
                    nsmap[prefix] = namespace

            wrapper = self.tree.parser.makeelement(
                utils.meta_attr('wrapper'), root.attrib.copy(), nsmap=nsmap)
            wrapper.append(root)
            root = wrapper

        # output XML headers, if applicable
        if global_scope is True or macro is "":
            header = ""
            if self.xml_declaration is not None:
                header += self.xml_declaration + '\n'
            if self.doctype:
                doctype = self.doctype + '\n'
                if self.encoding:
                    doctype = doctype.encode(self.encoding)
                header += doctype
            if header:
                out = clauses.Out(header)
                stream.scope.append(set())
                stream.begin([out])
                stream.end([out])
                stream.scope.pop()

        # add meta settings
        root.meta_omit_default_prefix = self.omit_default_prefix

        # start generation
        root.start(stream)
        body = stream.getvalue()

        # symbols dictionary
        __dict__ = stream.symbols.as_dict()

        # prepare globals
        _globals = ["from cPickle import loads as _loads"]
        for symbol, value in stream.symbol_mapping.items():
            _globals.append(
                "%s = _loads(%s)" % (symbol, repr(dumps(value))))

        transient = []
        for name, value in stream.symbols.as_dict().items():
            if value is config.TRANSIENT_SYMBOL:
                transient.append("%s = econtext.get('%s')" % (name, name))
        transient = "; ".join(transient)

        # wrap generated Python-code in function definition
        if global_scope:
            source = generation.function_wrap(
                'render', _globals, body, transient,
                "%(out)s.getvalue()" % __dict__)
        else:
            source = generation.function_wrap(
                'render', _globals, body, transient)

        suite = codegen.Suite(source)
        return suite.source



import config
import utils
import base64
import re

import xml.parsers.expat

from cPickle import dumps, loads
from StringIO import StringIO

re_tag_start = re.compile(r'<(?!!)')
marker = object()

def parse(body, parser, validator):
    """Parse XML document using expat and return an
    ElementTree-compatible document tree."""
    
    junk = ""
    tree = None
    parts = []

    offset = 0
    original_body = body
    
    while tree is None:
        expat = xml.parsers.expat.ParserCreate()
        expatparser = ExpatParser(parser, body, expat, validator)
        
        try:
            # attempt to parse this body; if we're not successful,
            # this may be because the document source consists of
            # several 'parts'; although this is not valid XML, we do
            # support it, being a template engine, not a XML
            # validator :-)
            if body is not None:
                expat.Parse(body, 1)

            root = expatparser.root
            if parts:
                if root is not None:
                    parts.append(root)
                root = parser.makeelement(
                    utils.meta_attr('fragments'))
                for i, part in enumerate(parts):
                    if isinstance(part, basestring):
                        parts[i-1].tail = part
                    else:
                        root.append(part)
                tree = root.getroottree()
            elif root is not None:
                tree = root.getroottree()
            else:
                raise RuntimeError("Unexpected parsing error.")

        except xml.parsers.expat.ExpatError, e:
            offset += expat.CurrentByteIndex

            # if we are not able to find a tree root, we give up and
            # let the exception through; if the error status code is 3
            # (no element found) and we've already parsed some parts,
            # we ignore the error and stop parsing.
            code = getattr(e, 'code', -1)
            if code == 3:
                body = None

                if parts:
                    continue

                raise

            if code == 9 and expatparser.root is not None:
                # add the root as a tree fragment and update the body
                # source to the next possible chunk
                parts.append(expatparser.root)
                body = body[:expatparser.index] + body[expat.CurrentByteIndex:]

                # a simple heuristic is used here to allow chunks of
                # 'junk' in-between the tree fragments
                m = re_tag_start.search(body)
                if m is not None:
                    pos = m.start()
                else:
                    pos = -1
                junk = body[:pos]
                body = body[pos:]
                parts.append(junk)
                offset += pos
            else:
                # the default exception won't provide correct
                # information about where in the document the
                # exception stems from; we can compute it manually
                # though.
                line = original_body[:offset].count('\n') + 1
                column = offset - original_body[:offset].rfind('\n')
                error_msg = str(e).split(':')[0]
                error_msg = "%s: line %d, column %d" % (
                    error_msg, line, column)
                raise xml.parsers.expat.ExpatError(error_msg)

    return tree, expatparser.doctype

class ExpatParser(object):
    """XML tree parser using the ``xml.parsers.expat`` stream
    parser. This parser serve to accept template input which lacks a
    proper prefix namespace mapping or entity definitions. It also
    works around an issue where the expat parser incorrectly parses an
    element with trivial body text as self-closing."""
    
    root = None
    index = None
    
    # doctype
    _doctype = doctype = None

    # xml-declaration
    xml_version = None
    encoding = None
    standalone = None
    parsing_external_entities = False
    
    def __init__(self, parser, body, expat, validator):
        self.parser = parser
        self.body = body
        self.expat = expat
        self._validate = validator

        # set up expat parser
        expat.ordered_attributes = 1
        expat.UseForeignDTD()
        expat.SetParamEntityParsing(
            xml.parsers.expat.XML_PARAM_ENTITY_PARSING_ALWAYS)

        # attach expat parser methods
        for name, value in type(self).__dict__.items():
            if isinstance(value, property):
                continue
            try:
                setattr(expat, name, getattr(self, name))
            except AttributeError:
                pass

    def _get_element(self):
        try:
            return self._element
        except AttributeError:
            raise xml.parsers.expat.ExpatError(
                "Unable to parse document; no start-tag found.")

    def _set_element(self, element):
        self._element = element

    element = property(_get_element, _set_element)

    def _validate(self, element):
        pass
    
    def StartElementHandler(self, tag, attributes):
        attrs = utils.odict()
        while attributes:
            key = attributes.pop(0)
            value = attributes.pop(0)
            attrs[key] = value
        
        # update prefix to namespace mapping
        if self.root is None:
            self.index = self.expat.CurrentByteIndex
            nsmap = utils.odict()
        else:
            nsmap = self.root.nsmap.copy()

        # process namespace declarations
        for key, value in attrs.items():
            if key.startswith('xmlns:'):
                prefix, name = key.split(':')
                nsmap[name] = value
                del attrs[key]

        for key, value in attrs.items():
            try:
                prefix, name = key.split(':')
            except (ValueError, TypeError):
                continue

            del attrs[key]

            namespace = nsmap.get(prefix)
            if namespace is None:
                if self.root is not None:
                    element = self.element

                    while element is not None:
                        namespace = element.nsmap.get(prefix)
                        if namespace is not None:
                            nsmap[prefix] = namespace
                            break
                        element = element.getparent()

                if namespace is None:
                    try:
                        namespace = config.DEFAULT_NS_MAP[prefix]
                    except KeyError:
                        raise KeyError(
                            "Attribute prefix unknown: '%s'." % prefix)

            attrs['{%s}%s' % (namespace, name)] = value

        # process tag
        try:
            prefix, name = tag.split(':')
            namespace = nsmap.get(prefix) or config.DEFAULT_NS_MAP[prefix]
            tag = '{%s}%s' % (namespace, name)
        except ValueError:
            pass

        # create element using parser
        element = self.parser.makeelement(tag, attrs, nsmap=nsmap)

        if self.root is None:
            # cook doctype
            doctype = []
            if (self._doctype is not None and
                self._doctype["pubid"] is None and
                self._doctype["sysid"] is None and
                self._doctype["has_internal_subset"] and
                self._doctype["entities"]):
                doctype.append("<!DOCTYPE %(doctype_name)s [" % self._doctype)
                for entity, repl in self._doctype["entities"]:
                    doctype.append('<!ENTITY %s "%s">' % (
                        entity,
                        repl.encode("ascii", "xmlcharrefreplace")))
                doctype.append("]>")
            elif self._doctype:
                d = self._doctype
                if d['pubid'] is None or d['sysid'] is None:
                    doctype.append(
                        '<!DOCTYPE %(doctype_name)s>' % d)
                else:
                    doctype.append(
                        '<!DOCTYPE %(doctype_name)s PUBLIC '
                        '"%(pubid)s" "%(sysid)s">' % d)
            self.doctype = "\n".join(doctype).encode("utf-8")

            ElementTree(
                self.parser, element, self.xml_version,
                self.encoding, self.standalone,
                self.doctype)

            # set this element as tree root
            self.root = element
        else:
            self.element.append(element)

        # validate element
        self._validate(element)

        # set as current element
        self.element = element

    def EndElementHandler(self, name):
        if self.element.text is None and self.body[
            self.expat.CurrentByteIndex-2] != '/':
            self.element.text = ""
        self.element = self.element.getparent()

    def CharacterDataHandler(self, data):
        """Process ``data`` while comparing to input context to check
        for HTML entities (which must be preserved)."""

        
        context = self.body[self.expat.CurrentByteIndex:]
        text = u""

        while data:
            m = utils.re_entity.search(context)
            if m is None or m.start() >= len(data):
                text += data
                break

            n = utils.re_entity.match(data)
            if n is not None:
                length = n.end()
            else:
                length = 1

            text += context[0:m.end()]
            context = context[m.end():]
            data = data[m.start()+length:]
            
        if len(self.element) == 0:
            current = self.element.text or ""
            self.element.text = current + text
        else:
            current = self.element[-1].tail or ""
            self.element[-1].tail = current + text
            
    def ProcessingInstructionHandler(self, target, data):
        self.element.append(
            self.parser.makeprocessinginstruction(target, data))

    def StartCdataSectionHandler(self):
        element = self.parser.makeelement(
            utils.xhtml_attr('cdata'))
        element.meta_cdata = ""
        self.element.append(element)
        self.element = element            

    def EndCdataSectionHandler(self):
        self.element = self.element.getparent()

    def CommentHandler(self, text):
        if self.root is not None and self.element is not None:
            self.element.append(
                self.parser.makecomment(text))

    def XmlDeclHandler(self, xml_version, encoding, standalone):
        self.xml_version = xml_version
        self.encoding = encoding

        if standalone:
            self.standalone = 'yes'
        else:
            self.standalone = 'no'
        
    def ExternalEntityRefHandler(self, context, base, sysid, pubid):
        self.parsing_external_entities = True
        try:
            parser = self.expat.ExternalEntityParserCreate(context)
            parser.ProcessingInstructionHandler = self.ProcessingInstructionHandler
            parser.ParseFile(StringIO(utils.entities))
            return 1
        finally:
            self.parsing_external_entities = False

    def DefaultHandler(self, userdata):
        if userdata.startswith('&'):
            return self.CharacterDataHandler(userdata)            
                
    def StartDoctypeDeclHandler(self, *args):
        doctype_name, sysid, pubid, has_internal_subset = args
        self.has_internal_subset = has_internal_subset
        self._doctype = {"doctype_name" : doctype_name,
                         "sysid": sysid,
                         "pubid": pubid,
                         "has_internal_subset": has_internal_subset,
                         "entities": []}

    def EntityDeclHandler(self, *args):
        if self.parsing_external_entities:
            return
        if self._doctype is not None and self._doctype["has_internal_subset"]:
            (entityName, is_parameter_entity, value,
             base, systemId, publicId, notationName) = args
            if value is not None:
                self._doctype["entities"].append((entityName, value))

class Parser(object):
    element_mapping = utils.emptydict
    fallback = None
    
    def parse_xml(self, body):
        return self.parse(body)
    
    def parse_text(self, text):
        tree = self.parse('<meta xmlns="%s"></meta>' % config.META_NS)
        tree.getroot().text = text
        tree.getroot().meta_omit = ""
        tree.getroot().meta_structure = True
        return tree

    def parse(self, body):
        lookup = ElementNamespaceClassLookup(
            fallback=ElementDefaultClassLookup(self.fallback))
        for key, mapping in self.element_mapping.items():
            lookup.get_namespace(key).update(mapping)

        parser = XMLParser(resolve_entities=False, strip_cdata=False)
        parser.setElementClassLookup(lookup)

        tree, self.doctype = parse(body, parser, self.validate)
        return tree

    def validate(self, element):
        pass

    # for backwards-compatibility, the parser object must be callable
    # which invokes the ``parse``-method
    def __call__(self, body):
        return self.parse(body)

class Annotation(property):
    def __init__(self, name, default=None):
        property.__init__(self, self._get, self._set)
        self.name = name
        self.default = default

    def _get(instance, element):
        value = element.attrib.get(instance.name)
        if value is None:
            return instance.default
        if value:
            value = loads(base64.decodestring(value))
        return value

    def _set(instance, element, value):
        element.attrib[instance.name] = base64.encodestring(dumps(value))

class ElementBase(object):
    """ElementTree-compatible base element."""

    node = tree = parent = prefix = tail = text = None

    def __init__(self, tag, attrib=None, nsmap=None):
        if attrib is None:
            attrib = {}
        if nsmap is None:
            nsmap = {}

        self.children = []
        self.tag = tag
        self.attrib = attrib
        self.nsmap = nsmap

        if '}' in tag:
            namespace = tag[1:].split('}')[0]
        else:
            namespace = attrib.get('xmlns')

        for prefix, ns in nsmap.items() + config.DEFAULT_NS_MAP.items():
            if namespace == ns:
                self.prefix = prefix
                break

    def append(self, element):
        self.children.append(element)
        self.register(element)

    def insert(self, index, element):
        self.children.insert(index, element)
        self.register(element)

    def remove(self, element):
        self.children.remove(element)
        element.parent = element.tree = None

    def register(self, element):
        element.parent = self
        
        # if we're appending an element rooted to an element tree,
        # we'll let the current element take over that tree.
        if self.tree is None and element.tree is not None:
            self.tree = element.tree
            self.tree.root = self
            
        # inherit tree reference
        element.tree = self.tree

    def index(self, element):
        return self.children.index(element)

    def getchildren(self):
        return self.children
    
    def getparent(self):
        return self.parent

    def getroottree(self):
        if self.tree is None:
            parent = self.getparent()
            if parent is None:
                raise RuntimeError("Unable to locate root; element has no parent.")
            return parent.getroottree()
        return self.tree

    @property
    def makeelement(self):
        tree = self.getroottree()
        return tree.parser.makeelement

    def tag(self):
        return utils.serialize(self)

    def walk(self):
        yield self
        for child in self:
            for element in child.walk():
                yield element

    def __len__(self):
        return len(self.children)
    
    def __iter__(self):
        return iter(self.children)

    def __getitem__(self, key):
        return self.children[key]

    def __str__(self):
        return str(unicode(self))

    def __unicode__(self):
        return self.tag()

class ProcessingInstructionElement(ElementBase):
    attrib = {}
    children = ()
    
    def __init__(self, target, data):
        self.target = target
        self.data = data

    def tag(self):
        return u"<?%s %s?>" % (self.target, self.data)

class CommentElement(ElementBase):
    """XML comment.

    This element is modelled after the ``lxml`` comment element; the
    ``tag`` attribute should be a method that renders the comment.
    """
    
    tail = None
    attrib = {}
    children = ()
    
    def __init__(self, text):
        self.text = text

    def tag(self):
        return u"<!--%s-->" % self.text

class ElementTree(object):
    """ElementTree-compatible tree element."""

    def __init__(self, parser, root, xml_version=None, encoding=None,
                 standalone=None, doctype=None):
        if doctype is None:
            doctype = {}

        self.docinfo = DocInfo(xml_version, encoding, standalone, doctype)
        self.parser = parser
        self.root = root
        
        root.tree = self
        
    def getroot(self):
        return self.root
        
class XMLParser(object):
    """XML parser class."""
    
    def __init__(self, resolve_entities=True, strip_cdata=True):
        self.resolve_entities = resolve_entities
        self.strip_cdata = strip_cdata

    def setElementClassLookup(self, lookup):
        self.lookup = lookup

    @property
    def makeelement(self):
        return self.lookup.makeelement

    @property
    def makeprocessinginstruction(self):
        return ProcessingInstructionElement

    @property
    def makecomment(self):
        return CommentElement

class ElementNamespaceClassLookup(object):
    def __init__(self, fallback=None):
        self.fallback = fallback
        self.by_namespace = {}
        
    def get_namespace(self, key):
        return self.by_namespace.setdefault(key, {})

    def makeelement(self, tag, attrs=None, nsmap=None):
        if '}' in tag:
            namespace, name = tag[1:].split('}')
        else:
            namespace, name = None, tag
            
        mapping = self.get_namespace(namespace)
        
        element_class = mapping.get(name) or mapping.get(None)
        if element_class is not None:
            return element_class(tag, attrs, nsmap=nsmap)

        return self.fallback.makeelement(tag, attrs, nsmap=nsmap)

class ElementDefaultClassLookup(object):
    def __init__(self, element_class):
        self.makeelement = element_class

class DocInfo(object):
    """XML document information class."""

    def __init__(self, xml_version, encoding, standalone, doctype):
        self.xml_version = xml_version
        self.encoding = encoding
        self.standalone = standalone
        self.doctype = doctype



import optparse
import os
import re
import sys

try:
    from lxml import etree
except ImportError:
    etree = None

HTML_COMMENT = re.compile(r"<!--.*?-->")
EXPANSION = re.compile(r"\${.*?}")

NSMAP = { "xhtml" : "http://www.w3.org/1999/xhtml",
          "py"    : "http://genshi.edgewall.org/",
          "tal"   : "http://xml.zope.org/namespaces/tal",
          "i18n"  : "http://xml.zope.org/namespaces/i18n",
        }

def qn(ns, tag):
    return "{%s}%s" % (NSMAP[ns], tag)

ATTR_TRANSLATE = qn("i18n", "translate")
ATTR_NAME = qn("i18n", "name")
ATTR_ATTRIBUTES = qn("i18n", "attributes")
NO_I18N_TAGS = set([qn("xhtml", "script"),
                    qn("xhtml", "object")])
NO_I18N_ATTRS = set([qn("tal", "replace"),
                     qn("tal", "content"),
                     qn("py", "replace"),
                     qn("py", "content"),
                     ])
I18N_ATTRS = set(["title", "alt"])




class Counter(object):
    def __init__(self):
        self.count=1
    def __add__(self, other):
        self.count+=other
    def __int__(self):
        return self.count
    def __str__(self):
        return str(self.count)

def hasText(txt):
    if not txt:
        return False
    txt=HTML_COMMENT.sub(u"", txt)
    txt=EXPANSION.sub(u"", txt)
    return bool(txt.strip())
        

def mustTranslate(el):
    for attr in el.attrib.keys():
        if attr in NO_I18N_ATTRS:
            return False

    if hasText(el.text):
        return True

    for child in el:
        if hasText(child.tail):
            return True

    return False


def handleContent(el, counter):
    if el.tag in NO_I18N_TAGS or not mustTranslate(el):
        if ATTR_TRANSLATE in el.attrib:
            del el.attrib[ATTR_TRANSLATE]
        return

    if ATTR_TRANSLATE in el.attrib:
        return

    el.attrib[ATTR_TRANSLATE]="string%d" % counter
    counter+=1

    childcounter=1
    for child in el:
        if ATTR_NAME in child.attrib:
            continue
        child.attrib[ATTR_NAME]="sub%d" % childcounter
        childcounter+=1


def getTranslatedAttributes(el):
    spec=el.attrib.get(ATTR_ATTRIBUTES, "")
    if ";" not in spec:
        return dict([(a,a) for a in spec.split()])
    else:
        return dict([a.split() for a in filter(None, spec.split(";"))])


def setTranslatedAttributes(el, mapping):
    if not mapping:
        if ATTR_ATTRIBUTES in el.attrib:
            del el.attrib[ATTR_ATTRIBUTES]
        return
    el.attrib[ATTR_ATTRIBUTES]=";".join("%s %s" % a for a in mapping.items())



def handleAttributes(el, counter):
    spec=getTranslatedAttributes(el)
    for attr in spec.keys():
        if not el.attrib.get(attr, "").strip():
            del spec[attr]
    for attr in I18N_ATTRS:
        if attr not in spec and el.attrib.get(attr, "").strip():
            spec[attr]="string%s" % counter
            counter+=1
    setTranslatedAttributes(el, spec)


def main():
    parser=optparse.OptionParser(
            usage="i18nify [options] [<template>..]",
            description="This script adds i18n attributes to Genshi and ZPT "
                        "templates so they can be translated by Chameleon.")
    parser.add_option("-o", "--output", metavar="FILE",
            help="Store the result in FILE instead.",
            dest="output", action="store")
    parser.add_option("-i", "--inplace",
            help="Modify template files in-place.",
            dest="inplace", action="store_true", default=False)

    if etree is None:
        print >>sys.stderr, "Must have ``lxml`` package installed."
        sys.exit(1)

    (options, args) = parser.parse_args()
    if not args:
        args=[sys.stdin]
    if options.output and len(args)>1:
        print >>sys.stderr, "Can not use -o/--output with multipe input files."
        sys.exit(1)
    if options.output and options.inplace:
        print >>sys.stderr, "Using both -i/--inplace and -o/--output is not possible."
        sys.exit(1)

    for filename in args:
        try:
            tree=etree.parse(filename)
        except (IOError, etree.XMLSyntaxError), e:
            msg=str(e)
            if len(args)>1:
                print >> sys.stderr, "%s: %s" % (filename, msg)
            else:
                print >> sys.stderr, msg
            sys.exit(2)
        root=tree.getroot()
        if "i18n" not in root.nsmap:
            newroot=etree.Element(root.tag,
                    nsmap=dict(i18n=NSMAP["i18n"], **root.nsmap))
            newroot[:]=root[:]
            root=newroot

        counter=Counter()
        for el in tree.iter():
            handleContent(el, counter)
            handleAttributes(el, counter)

        if options.inplace:
            tmpname="%s~" % filename
            tree.write(tmpname, encoding="utf-8")
            os.rename(tmpname, filename)
        else:
            if options.output:
                output=open(options.output, "w")
            else:
                output=sys.stdout

            if tree.docinfo.doctype:
                output.write(tree.docinfo.doctype)
                output.write("\n")
            output.write(etree.tostring(root))
            output.write("\n")


if __name__=="__main__":
    main()



import unittest
from chameleon.i18n import i18nize
from lxml.builder import ElementMaker

E=ElementMaker(namespace=i18nize.NSMAP["xhtml"],
                   nsmap={None: i18nize.NSMAP["xhtml"]})

class HasTextTests(unittest.TestCase):
    def testNone(self):
        self.assertEqual(i18nize.hasText(None), False)

    def testWhitespace(self):
        self.assertEqual(i18nize.hasText("  \t\n"), False)

    def testHtmlComment(self):
        self.assertEqual(i18nize.hasText("<!-- this is not relevant -->"), False)

    def testExpansion(self):
        self.assertEqual(i18nize.hasText("${myprecious}"), False)

    def testPlainText(self):
        self.assertEqual(i18nize.hasText("myprecious"), True)

    def testTextAndExpansion(self):
        self.assertEqual(i18nize.hasText("myprecious ${ring}"), True)


class MustTranslateTests(unittest.TestCase):
    def testEmptyElement(self):
        self.assertEqual(i18nize.mustTranslate(E.div()), False)

    def testElementWithText(self):
        self.assertEqual(i18nize.mustTranslate(E.p("some text")), True)

    def testElementWithChild(self):
        self.assertEqual(i18nize.mustTranslate(E.div(E.span())), False)

    def testElementWithTextAfterChild(self):
        span=E.span()
        span.tail="some text"
        self.assertEqual(i18nize.mustTranslate(E.div(span)), True)

    def testGenshiReplacedContent(self):
        self.assertEqual(i18nize.mustTranslate(E.div({i18nize.qn("py", "content") : "var"})), False)
        self.assertEqual(i18nize.mustTranslate(E.div({i18nize.qn("py", "replace") : "var"})), False)

    def testTalReplacedContent(self):
        self.assertEqual(i18nize.mustTranslate(E.div({i18nize.qn("tal", "content") : "var"})), False)
        self.assertEqual(i18nize.mustTranslate(E.div({i18nize.qn("tal", "replace") : "var"})), False)



class HandleContentTests(unittest.TestCase):
    def testNoText(self):
        el=E.span()
        i18nize.handleContent(el, None)
        self.assertEqual(el.attrib, {})

    def testAddTranslate(self):
        el=E.span("some text")
        i18nize.handleContent(el, i18nize.Counter())
        self.assertEqual(el.attrib, {i18nize.ATTR_TRANSLATE: "string1"})

    def testRemoveUnneededTranslate(self):
        el=E.span({i18nize.ATTR_TRANSLATE: "name"})
        i18nize.handleContent(el, None)
        self.assertEqual(el.attrib, {})

    def testNameChild(self):
        el=E.span("some ", E.strong("text"))
        i18nize.handleContent(el, i18nize.Counter())
        self.assertEqual(el.attrib, {i18nize.ATTR_TRANSLATE: "string1"})
        self.assertEqual(el[0].attrib, {i18nize.ATTR_NAME: "sub1"})

    def testKeepExistingTranslate(self):
        el=E.span("some text", {i18nize.ATTR_TRANSLATE: "name"})
        i18nize.handleContent(el, None)
        self.assertEqual(el.attrib, {i18nize.ATTR_TRANSLATE: "name"})

    def testKeepExistingChildName(self):
        el=E.span("some ", E.strong("text", {i18nize.ATTR_NAME: "text"}))
        i18nize.handleContent(el, i18nize.Counter())
        self.assertEqual(el.attrib, {i18nize.ATTR_TRANSLATE: "string1"})
        self.assertEqual(el[0].attrib, {i18nize.ATTR_NAME: "text"})


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



from chameleon.ast.astutil import parse
from chameleon.ast.astutil import _ast as ast
from chameleon.ast.astutil import ASTTransformer
from chameleon.ast.astutil import ASTCodeGenerator
from chameleon.core import config

import __builtin__

CONSTANTS = frozenset(['False', 'True', 'None', 'NotImplemented', 'Ellipsis'])
SYMBOLS = config.SYMBOLS.as_dict().values()
UNDEFINED = object()
OP_IGNORE = 'OP_IGNORE'

def flatten(list):
    """Flattens a potentially nested sequence into a flat list.
    """
    l = []
    for elt in list:
        t = type(elt)
        if t is set or t is tuple or t is list or t is frozenset:
            for elt2 in flatten(elt):
                l.append(elt2)
        else:
            l.append(elt)
    return l

def lookup_attr(obj, key):
    try:
        return getattr(obj, key)
    except AttributeError, e:
        try:
            get = obj.__getitem__
        except AttributeError:
            raise e
        try:
            return get(key)
        except KeyError:
            raise e

def lookup_name(data, name):
    try:
        return data[name]
    except KeyError:
        raise NameError(name)

lookup_globals = {
    '_lookup_attr': lookup_attr,
    '_lookup_name': lookup_name,
    }

class TemplateASTTransformer(ASTTransformer):
    def __init__(self, globals):
        self.locals = [CONSTANTS]
        builtin = dir(__builtin__)
        self.locals.append(set(globals))
        self.locals.append(set(builtin))
        # self.names is an optimization for visitName (so we don't
        # need to flatten the locals every time it's called)
        self.names = set()
        self.names.update(CONSTANTS)
        self.names.update(builtin)
        self.names.update(globals)

    def visit_Assign(self, node):
        node.value = self.visit(node.value)
        return ASTTransformer.visit_Assign(self, node)

    def visit_Delete(self, node):
        ASTTransformer.visit_Delete(self, node)

        # drop node
        pass

    def visit_FunctionDef(self, node):
        if len(self.locals) > 1:
            self.locals[-1].add(node.name)
            self.names.add(node.name)

        # process defaults *before* defining parameters
        node.args.defaults = tuple(
            self.visit(x) for x in node.args.defaults)

        if node.args.args:
            argnames = [arg.id for arg in node.args.args]
            self.locals.append(set(argnames))
            argnames = set(argnames)
            newnames = argnames.difference(self.names)
            self.names.update(newnames)

        try:
            return ASTTransformer.visit_FunctionDef(self, node)
        finally:
            if node.args.args:
                self.locals.pop()
                self.names -= newnames

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            if node.id not in self.names:
                # this should be spellable as...
                ##                     return ast.Subscript(
                ##                         ast.Name("econtext", ast.Load()),
                ##                         ast.Index(ast.Str(node.id)),
                ##                         ast.Load())
                ## .. but Python 2.5 doesn't allow it.
                name = ast.Name()
                name.id = 'econtext'
                name.ctx = ast.Load()
                idx = ast.Index()
                idx.value = ast.Str()
                idx.value.s = node.id
                subscr = ast.Subscript()
                subscr.value = name
                subscr.slice = idx
                subscr.ctx = ast.Load()
                return subscr


        if isinstance(node.ctx, (ast.Param, ast.Store)):
            self.locals[-1].add(node.id)
            self.names.add(node.id)

        if isinstance(node.ctx, ast.Del):
            self.locals[-1].remove(node.id)
            self.names.remove(node.id)

        return node

    def visit_ImportFrom(self, node):
        for index, alias in enumerate(node.names):
            if alias.asname is None:
                self.names.add(alias.name)
            else:
                self.names.add(alias.asname)
        return node

    def visit_Attribute(self, node):
        """Get attribute with fallback to dictionary lookup.

        Note: Variables starting with an underscore are exempt
        (reserved for internal use); as are the default system symbols.
        """

        if isinstance(node.value, ast.Name) and \
           (node.value.id.startswith('_') or node.value.id in SYMBOLS):
            return ASTTransformer.visit_Attribute(self, node)

        ## This should be spellable as
        ##         return ast.Call(
        ##             ast.Name('_lookup_attr', ast.Load()),
        ##             [self.visit(node.value), ast.Str(node.attr)],
        ##             [], None, None)
        ## .. except Python 2.5 doesn't allow it.

        call = ast.Call()
        name = ast.Name()
        name.id = '_lookup_attr'
        name.ctx = ast.Load()
        call.func = name
        string = ast.Str()
        string.s = node.attr
        args = [self.visit(node.value), string]
        call.args = args
        call.keywords = []
        call.starargs = None
        call.kwargs = None
        return call

    def visit_arguments(self, node):
        node.args = [self.visit(arg) for arg in node.args]
        return node

    def visit_Lambda(self, node):
        self.locals.append(set())
        try:
            return super(TemplateASTTransformer, self).visit_Lambda(node)
        finally:
            self.locals.pop()

    def visit_ListComp(self, node):
        self.locals.append(set())
        node.generators = [self.visit(gen) for gen in node.generators]
        node = super(TemplateASTTransformer, self).visit_ListComp(node)
        self.locals.pop()
        return node

class Suite(object):
    __slots__ = ['source', '_globals']

    mode = 'exec'

    def __init__(self, source, globals=()):
        """Create the code object from a string."""

        if isinstance(source, unicode):
            source = source.encode('utf-8')

        node = parse(source, self.mode)
        transform = TemplateASTTransformer(globals)
        tree = transform.visit(node)
        generator = ASTCodeGenerator(tree)
        self.source = generator.code

    def __hash__(self):
        return hash(self.source)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.source)



import os
import utils

# define which values are read as true
TRUEVALS = ('y', 'yes', 't', 'true', 'on', '1')

# in debug-mode, templates on disk are reloaded if they're modified
DEBUG_MODE_KEY = 'CHAMELEON_DEBUG'
DEBUG_MODE = os.environ.get(DEBUG_MODE_KEY, 'false')
DEBUG_MODE = DEBUG_MODE.lower() in TRUEVALS

# disable disk-cache to prevent the compiler from caching on disk
DISK_CACHE_KEY = 'CHAMELEON_CACHE'
DISK_CACHE = os.environ.get(DISK_CACHE_KEY, 'false')
DISK_CACHE = DISK_CACHE.lower() in TRUEVALS

# if eager parsing is enabled, templates are parsed upon
# instantiation, rather than when first called upon; this mode is
# useful for verifying validity of templates across a project
EAGER_PARSING_KEY = 'CHAMELEON_EAGER'
EAGER_PARSING = os.environ.get(EAGER_PARSING_KEY, 'false')
EAGER_PARSING = EAGER_PARSING.lower() in TRUEVALS

# in strict mode, filled macro slots must exist in the macro that's
# being used.
STRICT_MODE_KEY = 'CHAMELEON_STRICT'
STRICT_MODE = os.environ.get(STRICT_MODE_KEY, 'false')
STRICT_MODE = STRICT_MODE.lower() in TRUEVALS

# when validation is enabled, dynamically inserted content is
# validated against the XHTML standard
VALIDATION_KEY = 'CHAMELEON_VALIDATE'
VALIDATION = os.environ.get(VALIDATION_KEY, 'false')
VALIDATION = VALIDATION.lower() in TRUEVALS

# these definitions are standard---change at your own risk!
XML_NS = "http://www.w3.org/XML/1998/namespace"
XHTML_NS = "http://www.w3.org/1999/xhtml"
TAL_NS = "http://xml.zope.org/namespaces/tal"
META_NS = "http://xml.zope.org/namespaces/meta"
METAL_NS = "http://xml.zope.org/namespaces/metal"
XI_NS = "http://www.w3.org/2001/XInclude"
I18N_NS = "http://xml.zope.org/namespaces/i18n"
PY_NS = "http://genshi.edgewall.org/"

DEFAULT_ENCODING = "utf-8"

# default prefix namespace mapping
DEFAULT_NS_MAP = {
    None: XHTML_NS,
    'xml': XML_NS,
    'meta': META_NS,
    'py': PY_NS,
    'tal': TAL_NS,
    'metal': METAL_NS,
    'i18n': I18N_NS,
    'xi': XI_NS}

TRANSIENT_SYMBOL = object()

# the symbols table below is used internally be the compiler
class SYMBOLS(object):
    # internal use only
    init = '_init'
    callback = '_callback'
    slots = '_slots'
    metal = '_metal'
    include = '_include'
    macro = '_macro'
    out = '_out'
    tmp = '_tmp'
    write = '_write'
    mapping = '_mapping'
    result = '_result'
    marker = '_marker'
    domain = '_domain'
    i18n_context = '_i18n_context'
    attributes = '_attributes'
    negotiate = '_negotiate'
    translate = '_translate'
    validate = '_validate'
    msgid = '_msgid'
    re_amp = '_re_amp'
    raise_exc = '_raise_exc'

    # transient
    _slots = TRANSIENT_SYMBOL
    _translate = TRANSIENT_SYMBOL
    target_language = TRANSIENT_SYMBOL

    # markers
    default_marker = utils.default()
    default_marker_symbol = '_default'

    # advertised symbols
    repeat = 'repeat'
    language = 'target_language'
    xincludes = 'xincludes'
    default = 'default'
    scope = 'econtext'
    remote_scope = 'rcontext'

    def __new__(cls, **kwargs):
        class SYMBOLS(cls):
            pass

        for name, value in kwargs.items():
            setattr(SYMBOLS, name, value)

        return SYMBOLS

    @classmethod
    def as_dict(cls):
        return dict((name, getattr(cls, name)) for name in dir(cls))



# -*- encoding: utf-8 -*-
import unittest

class TestUnicodeMisc(unittest.TestCase):
    def test_utf8_in_page_text_template_file(self):
        import os
        path = os.path.join(os.path.dirname(__file__), "templates")
        from chameleon.zpt.template import PageTextTemplateFile
        t = PageTextTemplateFile(os.path.join(path, 'helloworld.txt'))
        result = t()
        self._assert_unicode_equals(
            result, u'Hello W\xf5rld!\n')

    def test_utf8_values_in_page_text_template_file(self):
        import os
        path = os.path.join(os.path.dirname(__file__), "templates")
        from chameleon.zpt.template import PageTextTemplateFile
        t = PageTextTemplateFile(os.path.join(path, 'uglyworld.txt'))
        value = u'ä'.encode("utf-8")
        result = t(value=value)
        self._assert_unicode_equals(
            result, u'Inserting a value: ä\n')

    def test_unicode_values_in_page_text_template_file(self):
        import os
        path = os.path.join(os.path.dirname(__file__), "templates")
        from chameleon.zpt.template import PageTextTemplateFile
        t = PageTextTemplateFile(os.path.join(path, 'uglyworld.txt'))
        value = u'ä'
        result = t(value=value)
        self._assert_unicode_equals(
            result, u'Inserting a value: ä\n')

    def _assert_unicode_equals(self, value, other):
        if not isinstance(value, unicode):
            try:
                value = value.decode('utf-8')
            except UnicodeEncodeError, e:
                self.fail(str(e))
        self.assertEqual(value, other)



from chameleon.core import template
from chameleon.genshi.language import MatchTemplates
from chameleon.genshi.language import Parser
from chameleon.genshi.language import TextParser

class GenshiTemplate(template.Template):
    __doc__ = template.Template.__doc__ # for Sphinx autodoc

    default_parser = Parser()

    def __init__(self, body, parser=None, **kwargs):
        if parser is None:
            parser = self.default_parser
        super(GenshiTemplate, self).__init__(body, parser, **kwargs)

    def render(self, *args, **kwargs):
        mt = kwargs['match_templates'] = MatchTemplates()
        result = super(GenshiTemplate, self).render(*args, **kwargs)
        return mt.process(result)

class GenshiTemplateFile(template.TemplateFile, GenshiTemplate):
    __doc__ = template.TemplateFile.__doc__ # for Sphinx autodoc

    def __init__(self, filename, parser=None, **kwargs):
        if parser is None:
            parser = self.default_parser
        super(GenshiTemplateFile, self).__init__(
            filename, parser, **kwargs)

class GenshiTextTemplate(GenshiTemplate):
    __doc__ = template.Template.__doc__ # for Sphinx autodoc
    default_parser = TextParser()
    format = 'text'

class GenshiTextTemplateFile(GenshiTemplateFile):
    __doc__ = template.Template.__doc__ # for Sphinx autodoc
    default_parser = TextParser()
    format = 'text'



try:
    from zope.i18n import interpolate
    from zope.i18n import translate
    from zope.i18nmessageid import Message
except ImportError:
    def fast_translate(msgid, *args, **kwargs):
        return kwargs.get('default', msgid)
else:
    def fast_translate(msgid, domain=None, mapping=None, context=None,
                       target_language=None, default=None):
        if msgid is None:
            return

        if target_language is not None:
            return translate(
                msgid, domain=domain, mapping=mapping, context=context,
                target_language=target_language, default=default)

        if isinstance(msgid, Message):
            default = msgid.default
            mapping = msgid.mapping

        if default is None:
            default = unicode(msgid)

        if not isinstance(default, basestring):
            return default

        return interpolate(default, mapping)

class StringMarker(str):
    def __nonzero__(self):
        return False

marker = StringMarker()




#!/usr/bin/env python # -- coding: utf-8 --

from setuptools import setup, find_packages
import sys

version = '1.2.0'

install_requires = [
    'setuptools',
    ]

if sys.version_info[:3] < (2,5,0):
    install_requires.append('elementtree')

setup(name="Chameleon",
      version=version,
      description="XML-based template compiler.",
      long_description=open("README.rst").read() + open("CHANGES.rst").read(),
      classifiers=[
        "Programming Language :: Python",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: XML",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='',
      author='Malthe Borch and the Repoze Community',
      author_email="repoze-dev@lists.repoze.org",
      url="http://chameleon.repoze.org",
      license='BSD-like (http://repoze.org/license.html)',
      namespace_packages=['chameleon'],
      packages = find_packages('src'),
      package_dir = {'':'src'},
      include_package_data=True,
      zip_safe=False,
      entry_points = """
      [console_scripts]
      i18nize = chameleon.i18n.i18nize:main [i18nize]
      """,
      install_requires=install_requires,
      extras_require = {
          'i18nize' : [ "lxml", ],
          },
      tests_require=install_requires + [
          'zope.interface==3.5.2',
          'zope.component==3.7.1',
          'zope.i18n==3.7.1',
          'lxml',
          ],
      test_suite="chameleon.tests",
      )



from chameleon.core import translation
from chameleon.core import generation
from chameleon.core import config
from chameleon.core import etree
from chameleon.core import utils
from chameleon.core import types

import expressions
import itertools

try:
    from lxml import html
except ImportError:
    html = None

def get_elements_with_attribute_unless_in_between(element, ns, name, discard):
    elements = []

    discarded = utils.elements_with_attribute(element, ns, discard)
    subelements = utils.elements_with_attribute(element, ns, name)

    for subelement in subelements:
        parent = subelement.getparent()
        while parent is not None:
            if parent in discarded:
                break
            parent = parent.getparent()
        else:
            elements.append(subelement)

    return elements

class XPathResult(list):
    def __unicode__(self):
        def format_result(result):
            if isinstance(result, basestring):
                return result
            return html.tostring(result)
        return u"".join(
            map(format_result, self))

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __html__(self):
        return unicode(self)

class MatchTemplates(object):
    encoding = 'utf-8'

    def __init__(self):
        self.templates = {}

    def __call__(self, path, limit):
        def outer(func):
            self.templates[path] = func, limit
            def inner(*args, **kwargs):
                return func(*args, **kwargs)
            return inner
        return outer

    def __nonzero__(self):
        return bool(self.templates)

    def process(self, body):
        if not self.templates:
            return body

        if html is None:
            raise ImportError("lxml.html")

        root = html.fromstring(body)
        for path, (method, limit) in self.templates.items():
            for element in root.xpath(path)[:limit]:
                out, write = generation.initialize_stream()
                def select(path):
                    return XPathResult(element.xpath(path))
                method(out, write, select)

                # replace element with fragments
                fragments = html.fragments_fromstring(out.getvalue())
                if element is root:
                    for fragment in fragments:
                        # ignore trivial fragments, if we're replacing
                        # the root node
                        if isinstance(fragment, basestring):
                            if fragment.strip('\n ') == "":
                                continue
                            raise ValueError(
                                "Must replace root with structural element.")

                        prev = root = fragment
                else:
                    # this node does have a parent; replace it with
                    # the fragments
                    tail = element.tail
                    prev = element.getprevious()
                    parent = element.getparent()
                    index = parent.index(element)

                    fragment = fragments[0]
                    if isinstance(fragment, basestring):
                        if prev is None:
                            parent.text = (parent.text or "") + fragment
                        else:
                            prev.tail = (prev.tail or "") + fragment
                        fragments.pop(0)

                    for fragment in fragments:
                        if isinstance(fragment, basestring):
                            assert prev is not None
                            prev.tail = (prev.tail or "") + fragment
                        else:
                            parent.insert(index+1, fragment)
                            prev = fragment

                    parent.remove(element)
                    if prev is None:
                        parent.text += tail
                    else:
                        prev.tail = tail

        return html.tostring(root, pretty_print=True, encoding=self.encoding)

class GenshiElement(translation.Element):
    """Genshi language element."""

    translator = expressions.translator
    strip_text = False

    class node(translation.Node):
        ns_omit = (
            "http://xml.zope.org/namespaces/meta",
            "http://xml.zope.org/namespaces/i18n",
            "http://www.w3.org/2001/XInclude",
            "http://genshi.edgewall.org/")

        match_symbol = "match_templates"
        markup_symbol = "_markup"

        symbols = translation.Node.symbols(
            match_templates=config.TRANSIENT_SYMBOL)

        @property
        def _interpolation_enabled(self):
            return self.element.meta_interpolation in config.TRUEVALS + ('',)

        @property
        def _interpolation_escape(self):
            return self.element.meta_interpolation_escaping in config.TRUEVALS + ('',)

        @property
        def omit(self):
            if self.element.py_strip is not None:
                return self.element.py_strip or True
            if self.element.meta_omit is not None:
                return self.element.meta_omit or True
            if self.element.py_replace or self.element.meta_replace:
                return True
            if self.element.xi_href:
                return True

        @property
        def define(self):
            return self.element.py_with

        @property
        def condition(self):
            return self.element.py_if

        @property
        def repeat(self):
            return self.element.py_for

        @property
        def content(self):
            return self.element.py_content or self.element.py_replace or \
                   self.element.meta_replace

        @property
        def skip(self):
            return bool(self.content) or self.translate is not None

        @property
        def dict_attributes(self):
            return self.element.py_attrs

        @property
        def dynamic_attributes(self):
            attributes = []

            py_attributes = utils.get_attributes_from_namespace(
                self.element, config.PY_NS)
            meta_attributes = utils.get_attributes_from_namespace(
                self.element, config.META_NS)
            i18n_attributes = utils.get_attributes_from_namespace(
                self.element, config.I18N_NS)

            internal = tuple(itertools.chain(
                py_attributes, meta_attributes, i18n_attributes))

            if self._interpolation_enabled:
                attributes.extend(self.interpolated_attributes(internal))

            if self.element.meta_attributes is not None:
                attributes.extend(self.element.meta_attributes)

            if len(attributes) > 0:
                return attributes

        @property
        def macro(self):
            if self.element.py_match is not None:
                if self.match_symbol not in itertools.chain(
                    *self.element.stream.scope):
                    raise NameError(self.match_symbol)
                symbols = self.element.stream.symbols
                args = symbols.out, symbols.write, "select"
                if self.element.py_once.lower() in config.TRUEVALS:
                    once = 1
                else:
                    once = repr(None)
                decorator = '%s("""%s""", %s)' % (
                    self.match_symbol, self.element.py_match, once)
                return types.method(
                    "match", args, decorators=(decorator,))

            return self.element.py_def

        @property
        def translated_attributes(self):
            return self.element.i18n_attributes

        @property
        def translate(self):
            return self.element.i18n_translate

        @property
        def translation_name(self):
            return self.element.i18n_name

        @property
        def translation_domain(self):
            return self.element.i18n_domain

        @property
        def cdata(self):
            return self.element.meta_cdata

        @property
        def include(self):
            href = self.element.xi_href
            if href is not None:
                return types.join(self.element.translator.split(href))

        @property
        def format(self):
            return self.element.xi_parse

        @property
        def text(self):
            text = self.element.text
            if text is not None:
                parent = self.element.getparent()
                if parent is not None and self.element.strip_text and parent[0] is self.element:
                    text = text.lstrip('\n ')

                if not self._interpolation_enabled:
                    return (text,)

                parts = self.element.translator.split(text)
                if self.element.tag == '{http://www.w3.org/1999/xhtml}cdata':
                    return parts

                if not self._interpolation_escape:
                    return parts

                return tuple(
                    isinstance(part, types.expression) and types.escape(part) or \
                    self.element.meta_structure is True and part or \
                    utils.htmlescape(part) for part in parts)

            return ()

        @property
        def tail(self):
            tail = self.element.tail
            if tail is not None:
                if self.element.strip_text:
                    parent = self.element.getparent()
                    if parent is not None and parent.strip_text and parent[-1] is self.element:
                        tail = tail.rstrip('\n ')

                parent = self.element.getparent()
                if parent is not None and not parent.node._interpolation_enabled:
                    return (tail,)

                parts = self.element.translator.split(tail)
                if self.element.tag == '{http://www.w3.org/1999/xhtml}cdata':
                    return parts

                if not self._interpolation_escape:
                    return parts

                return tuple(
                    isinstance(part, types.expression) and types.escape(part) or \
                    self.element.meta_structure is True and part or \
                    utils.htmlescape(part) for part in parts)

            return ()

    node = property(node)

    def update(self):
        # Step 1: Convert py:choose, py:when, py:otherwise into
        # tal:define, tal:condition
        stream = self.stream
        choose_expression = self._pull_attribute(utils.py_attr('choose'))
        if choose_expression is not None:
            choose_variable = stream.new_var()
            
            if choose_expression:
                self._add_define(choose_variable, choose_expression)
                
            # select all elements that have the "py:when" controller,
            # unless a "py:choose" expression sits in-between
            elements = get_elements_with_attribute_unless_in_between(
                self, config.PY_NS, "when", "choose")

            variables = []
            for element in elements:
                expression = element._pull_attribute(utils.py_attr('when'))
                variable = stream.new_var()
                variables.append(variable)

                # add definition to ancestor
                self._add_define(variable, expression)

                # add condition to element
                if choose_expression:
                    expression = "%s == %s" % (
                        choose_variable, variable)
                else:
                    expression = "%s" % variable

                element.attrib[utils.py_attr('if')] = expression

            # process any "py:otherwise"-controllers
            elements = get_elements_with_attribute_unless_in_between(
                self, config.PY_NS, "otherwise", "choose")

            for element in elements:
                if choose_expression:
                    expression = "%s not in %s" % (
                        choose_variable, repr(tuple(variables)))
                else:
                    expression = "not(%s)" % " or ".join(variables)

                element.attrib[utils.py_attr('if')] = expression

    def _add_define(self, variable, expression):
        name = utils.py_attr('with')
        define = "%s=%s; " % (variable, expression)

        if name in self.attrib:
            self.attrib[name] += define
        else:
            self.attrib[name] = define

    def _pull_attribute(self, name, default=None):
        attrib = self.attrib
        if name in attrib:
            value = attrib[name]
            del attrib[name]
            return value
        return default

    xi_href = None
    xi_parse = None
    py_once = "false"

class XHTMLElement(GenshiElement):
    """XHTML namespace element."""

    py_if = utils.attribute(
        utils.py_attr('if'), lambda p: p.expression)
    py_for = utils.attribute(
        utils.py_attr('for'), lambda p: p.definition)
    py_with = utils.attribute(utils.py_attr('with'),
        lambda p: expressions.translator.definitions)
    py_choose = utils.attribute(
        utils.py_attr('choose'), lambda p: p.expression)
    py_when = utils.attribute(
        utils.py_attr('when'), lambda p: p.expression)
    py_otherwise = utils.attribute(
        utils.py_attr('otherwise'), lambda p: p.expression)
    py_match = utils.attribute(
        utils.py_attr('match'))
    py_once = utils.attribute(
        utils.py_attr('once'), default='false')
    py_def = utils.attribute(
        utils.py_attr('def'), lambda p: p.method)
    py_attrs = utils.attribute(
        utils.py_attr('attrs'), lambda p: p.expression)
    py_content = utils.attribute(
        utils.py_attr('content'), lambda p: p.output)
    py_replace = utils.attribute(
        utils.py_attr('replace'), lambda p: p.output)
    py_strip = utils.attribute(
        utils.py_attr('strip'), lambda p: p.expression)
    meta_translator = etree.Annotation(
        utils.meta_attr('translator'))
    meta_interpolation = utils.attribute(
        utils.meta_attr('interpolation'), default='true', recursive=True)
    meta_interpolation_escaping = utils.attribute(
        utils.meta_attr('interpolation-escaping'), default='true', recursive=True)
    i18n_translate = utils.attribute(
        utils.i18n_attr('translate'))
    i18n_attributes = utils.attribute(
        utils.i18n_attr('attributes'), lambda p: p.mapping)
    i18n_domain = utils.attribute(
        utils.i18n_attr('domain'))
    i18n_name = utils.attribute(
        utils.i18n_attr('name'))

class MetaElement(XHTMLElement, translation.MetaElement):
    pass

class PyElement(XHTMLElement):
    strip_text = True
    py_strip = utils.attribute("strip", lambda p: p.expression, u"")

class PyIfElement(PyElement):
    py_if = utils.attribute("test", lambda p: p.expression)

class PyForElement(PyElement):
    py_for = utils.attribute("each", lambda p: p.definition)

class PyWithElement(PyElement):
    py_with = utils.attribute(
        "vars", lambda p: expressions.translator.definitions)

class PyDefElement(PyElement):
    py_def = utils.attribute("function", lambda p: p.method)

class PyMatchElement(PyElement):
    py_match = utils.attribute("path")
    py_once = utils.attribute("once", default="false")

class XiIncludeElement(XHTMLElement):
    xi_href = utils.attribute("href")
    xi_parse = utils.attribute("parse", default="xml")

class TextElement(XHTMLElement):
    meta_interpolation_escaping = False

class Parser(etree.Parser):
    """The parser implementation for Genshi templates."""

    element_mapping = {
        config.XHTML_NS: {None: XHTMLElement},
        config.META_NS: {None: MetaElement},
        config.XI_NS: {'include': XiIncludeElement},
        config.PY_NS: {'if': PyIfElement,
                       'for': PyForElement,
                       'def': PyDefElement,
                       'with': PyWithElement,
                       'match': PyMatchElement}}

    fallback = XHTMLElement

class TextParser(Parser):
    element_mapping = {
        config.XHTML_NS: {None: TextElement},
        }



try:
    from zope.interface import implements
except ImportError:
    def implements(*interfaces):
        pass

try:
    from zope.component import queryUtility
    from zope.component import queryAdapter
    from zope.component import adapts
except ImportError:
    def queryUtility(*args, **kw):
        return
    def queryAdapter(*args, **kw):
        return
    def adapts(*interfaces):
        pass

import sys
import interfaces
import htmlentitydefs
import re
import xml.parsers.expat

types = sys.modules['types']

from UserDict import UserDict

# check if we're able to coerce unicode to str
try:
    str(u'La Pe\xf1a')
    unicode_required_flag = False
except UnicodeEncodeError:
    unicode_required_flag = True

def coerces_gracefully(encoding):
    if encoding != sys.getdefaultencoding() and unicode_required_flag:
        return False
    return True

entities = "".join((
    '<!ENTITY %s "&amp;%s;">' % (name, name) for (name, text) in \
    htmlentitydefs.name2codepoint.items()))

re_annotation = re.compile(r'^\s*u?[\'"](.*)[\'"]$')
re_entity = re.compile(r'&([A-Za-z]+|#[0-9]+);')
re_numeric_entity = re.compile(r'&#([0-9]+);')
re_amp = re.compile(r'&(?!([A-Za-z]+|#[0-9]+);)');
re_meta = re.compile(
    r'\s*<meta\s+http-equiv=["\']?Content-Type["\']?'
    r'\s+content=["\']?([^;]+);\s*charset=([^"\']+)["\']?\s*/?\s*>\s*',
    re.IGNORECASE)

XML_PREFIXES = [
    "<?xml",                      # ascii, utf-8
    "\xef\xbb\xbf<?xml",          # utf-8 w/ byte order mark
    "\0<\0?\0x\0m\0l",            # utf-16 big endian
    "<\0?\0x\0m\0l\0",            # utf-16 little endian
    "\xfe\xff\0<\0?\0x\0m\0l",    # utf-16 big endian w/ byte order mark
    "\xff\xfe<\0?\0x\0m\0l\0",    # utf-16 little endian w/ byte order mark
    ]

XML_PREFIX_MAX_LENGTH = max(map(len, XML_PREFIXES))

s_counter = 0
marker = object()

class callablestr(str):
    __slots__ = ()

    def __call__(self):
        return self

class callableint(int):
    __slots__ = ()

    def __call__(self):
        return self

class descriptorstr(object):
    __slots__ = "function", "__name__"

    def __init__(self, function):
        self.function = function
        self.__name__ = function.__name__

    def __get__(self, context, cls):
        return callablestr(self.function(context))

class descriptorint(object):
    __slots__ = "function", "__name__"

    def __init__(self, function):
        self.function = function
        self.__name__ = function.__name__

    def __get__(self, context, cls):
        return callableint(self.function(context))

def sniff_type(text):
    """Return 'text/xml' if text appears to be XML, otherwise return None."""
    for prefix in XML_PREFIXES:
        if text.startswith(prefix):
            return "text/xml"
    return None

def handler(key=None):
    def decorate(f):
        def g(node):
            if key is None:
                return f(node, None)
            return f(node, node.get(key))
        g.__ns__ = key
        return g
    return decorate

def import_elementtree():
    """The ElementTree is used to validate output in debug-mode. We
    attempt to load the library from several locations."""
    
    try:
        import xml.etree.ElementTree as ET
    except:
        try:
            import elementtree.ElementTree as ET
        except ImportError:
            import cElementTree as ET
        
    return ET

def validate(string):
    """Wraps string as a proper HTML document and validates it by
    attempting to parse it using the active ElementTree parser."""
    
    validation_string = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd" [ %s ]><div>%s</div>' % (
        entities, string)
    
    try:
        import_elementtree().fromstring(validation_string)
    except xml.parsers.expat.ExpatError:
        raise ValidationError(string)

def elements_with_attribute(element, ns, name, value=None):
    """Returns elements on the 'descendant-or-self' axis that has
    ``attribute`` (in qualified namespace notation).

    If ``value`` is specified, only elemmnts with this attribute value
    are returned.
    """

    return tuple(_elements_with_attribute(element, ns, name, value))

def _elements_with_attribute(element, ns, name, value):
    element_ns = get_namespace(element)
    if element_ns == ns:
        _value = element.attrib.get(name, marker)
    else:
        _value = marker

    if _value is marker:
        _value = element.attrib.get('{%s}%s' % (ns, name), marker)

    if value is not None:
        if value == _value:
            yield element
    elif _value is not marker:
        yield element

    for child in element:
        for match in _elements_with_attribute(child, ns, name, value):
            yield match

class attribute(property):
    def __init__(self, tokens, factory=None, default=None, encoding=None, recursive=False):
        self.tokens = tokens
        self.factory = factory
        self.default = default
        self.encoding = encoding
        self.recursive = recursive
        property.__init__(self, self.__call__)

    def __call__(attribute, element):
        value = attribute.get(element)
        if value is not None:
            encoding = attribute.encoding
            if encoding is None:
                try:
                    encoding = element.stream.encoding
                except AttributeError:
                    pass
            if encoding:
                value = value.encode(encoding, 'xmlcharrefreplace')
            if attribute.factory is None:
                return value
            f = attribute.factory(element.translator)
            return f(value)

        if attribute.recursive:
            parent = element.getparent()
            if parent is not None:
                return attribute(parent)

        if attribute.default is not None:
            return attribute.default

    def get(attribute, element):
        tokens = attribute.tokens
        for token in isinstance(tokens, tuple) and tokens or (tokens,):
            value = element.attrib.get(token)
            if value is not None:
                return value

def escape(string, quote=None, encoding=None):
    if not isinstance(string, unicode) and encoding:
        string = string.decode(encoding)
    else:
        encoding = None

    string = re_amp.sub('&amp;', string).\
             replace('<', '&lt;').\
             replace('>', '&gt;')

    if quote is not None:
        string = string.replace(quote, '\\'+quote)

    if encoding:
        string = string.encode(encoding)

    return string

def htmlescape(text):
    while text:
        m = re_amp.search(text)
        if m is None:
            break
        text = text[:m.start()] + re_amp.sub("&amp;", text[m.start():], 1)

    return text.replace('<', '&lt;').replace('>', '&gt;')

re_normalize = re.compile(r'(^[0-9]|\b[^A-Za-z])')
def normalize_slot_name(name):
    return re_normalize.sub('_', name)

def serialize(element, encoding=None, omit=False):
    return "".join(serialize_element(element, encoding, omit=omit))

def serialize_element(element, encoding, omit):
    try:
        name = element.tag.split('}')[-1]
    except AttributeError:
        yield htmlescape(element.text); return

    if omit is False:
        # tag opening
        yield "<%s" % name

        # attributes
        for key, value in element.attrib.items():
            yield ' %s="%s"' % (key, escape('"', encoding=encoding))

        # elements with no text which have no children are self-closing.
        if element.text is None and len(element) == 0:
            yield ' />'; return

        yield '>'
            
    if element.text is not None:
        yield htmlescape(element.text)

    for child in element:
        for string in serialize_element(child, encoding, False):
            yield string

    if omit is False:
        yield '</%s>' % name

def class_hierarchy(cls):
    """Return an unordered sequence of all classes related to cls.

    Traverses diamond hierarchies.

    Fibs slightly: subclasses of builtin types are not returned.  Thus
    class_hierarchy(class A(object)) returns (A, object), not A plus every
    class systemwide that derives from object.

    Old-style classes are discarded and hierarchies rooted on them
    will not be descended.

    """
    if isinstance(cls, types.ClassType):
        return list()
    hier = set([cls])
    process = list(cls.__mro__)
    while process:
        c = process.pop()
        if isinstance(c, types.ClassType):
            continue
        for b in (_ for _ in c.__bases__
                  if _ not in hier and not isinstance(_, types.ClassType)):
            process.append(b)
            hier.add(b)
        if c.__module__ == '__builtin__' or not hasattr(c, '__subclasses__'):
            continue
        for s in [_ for _ in c.__subclasses__() if _ not in hier]:
            process.append(s)
            hier.add(s)
    return list(hier)

class default(object):
    __slots__ = 'value',

    def __init__(self):
        self.value = None

    def __setitem__(self, key, value):
        self.value = value

    def __html__(self):
        return self.value

class econtext(dict):
    """Dynamic scope dictionary which is compatible with the
    `econtext` of ZPT."""

    set_local = setLocal = dict.__setitem__
    set_global = setGlobal = dict.__setitem__

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            raise NameError(key)

    @property
    def vars(self):
        return self

    def copy(self):
        inst = econtext(self)
        inst.set_global = self.set_global
        return inst

class scope(list):
    def __init__(self, *args):
        global s_counter
        self.hash = s_counter
        s_counter += 1

    def __hash__(self):
        return self.hash

class emptydict(dict):
    def __setitem__(self, key, value):
        raise RuntimeError("Read-only dictionary does not support assignment.")

    def update(self, dictionary):
        raise RuntimeError("Read-only dictionary does not support assignment.")

emptydict = emptydict()

class repeatitem(object):
    implements(interfaces.ITALESIterator)

    __slots__ = "length", "_iterator"

    def __init__(self, iterator, length):
        self.length = length
        self._iterator = iterator

    try:
        iter(()).__len__
    except AttributeError:
        @property
        def index(self):
            remaining = self._iterator.__length_hint__()
            return self.length - remaining - 1
    else:
        @property
        def index(self):
            remaining = self._iterator.__len__()
            return self.length - remaining - 1

    @property
    def start(self):
        return self.index == 0

    @property
    def end(self):
        return self.index == self.length - 1

    @descriptorint
    def number(self):
        return self.index + 1

    @descriptorstr
    def odd(self):
        """Returns a true value if the item index is odd.

        >>> it = repeatitem(iter(("apple", "pear")), 2)

        >>> it._iterator.next()
        'apple'
        >>> it.odd()
        ''

        >>> it._iterator.next()
        'pear'
        >>> it.odd()
        'odd'
        """

        return self.index % 2 == 1 and 'odd' or ''

    @descriptorstr
    def even(self):
        """Returns a true value if the item index is even.

        >>> it = repeatitem(iter(("apple", "pear")), 2)

        >>> it._iterator.next()
        'apple'
        >>> it.even()
        'even'

        >>> it._iterator.next()
        'pear'
        >>> it.even()
        ''
        """

        return self.index % 2 == 0 and 'even' or ''

    def next(self):
        raise NotImplementedError(
            "Method not implemented (can't update local variable).")

    def _letter(self, base=ord('a'), radix=26):
        """Get the iterator position as a lower-case letter

        >>> it = repeatitem(iter(("apple", "pear", "orange")), 3)
        >>> it._iterator.next()
        'apple'
        >>> it.letter()
        'a'
        >>> it._iterator.next()
        'pear'
        >>> it.letter()
        'b'
        >>> it._iterator.next()
        'orange'
        >>> it.letter()
        'c'
        """

        index = self.index
        if index < 0:
            raise TypeError("No iteration position")
        s = ""
        while 1:
            index, off = divmod(index, radix)
            s = chr(base + off) + s
            if not index:
                return s

    letter = descriptorstr(_letter)

    @descriptorstr
    def Letter(self):
        """Get the iterator position as an upper-case letter

        >>> it = repeatitem(iter(("apple", "pear", "orange")), 3)
        >>> it._iterator.next()
        'apple'
        >>> it.Letter()
        'A'
        >>> it._iterator.next()
        'pear'
        >>> it.Letter()
        'B'
        >>> it._iterator.next()
        'orange'
        >>> it.Letter()
        'C'
        """

        return self._letter(base=ord('A'))

    @descriptorstr
    def Roman(self, rnvalues=(
                    (1000,'M'),(900,'CM'),(500,'D'),(400,'CD'),
                    (100,'C'),(90,'XC'),(50,'L'),(40,'XL'),
                    (10,'X'),(9,'IX'),(5,'V'),(4,'IV'),(1,'I')) ):
        """Get the iterator position as an upper-case roman numeral

        >>> it = repeatitem(iter(("apple", "pear", "orange")), 3)
        >>> it._iterator.next()
        'apple'
        >>> it.Roman()
        'I'
        >>> it._iterator.next()
        'pear'
        >>> it.Roman()
        'II'
        >>> it._iterator.next()
        'orange'
        >>> it.Roman()
        'III'
        """

        n = self.index + 1
        s = ""
        for v, r in rnvalues:
            rct, n = divmod(n, v)
            s = s + r * rct
        return s

    @descriptorstr
    def roman(self):
        """Get the iterator position as a lower-case roman numeral

        >>> it = repeatitem(iter(("apple", "pear", "orange")), 3)
        >>> it._iterator.next()
        'apple'
        >>> it.roman()
        'i'
        >>> it._iterator.next()
        'pear'
        >>> it.roman()
        'ii'
        >>> it._iterator.next()
        'orange'
        >>> it.roman()
        'iii'
        """

        return self.Roman().lower()

class repeatdict(dict):
    __slots__ = ()

    def insert(self, key, iterable):
        """We coerce the iterable to a tuple and return an iterator
        after registering it in the repeat dictionary."""

        try:
            iterable = tuple(iterable)
        except TypeError:
            if iterable is None:
                iterable = ()
            else:
                # The message below to the TypeError is the Python
                # 2.5-style exception message. Python 2.4.X also
                # raises a TypeError, but with a different message.
                # ("TypeError: iteration over non-sequence").  The
                # Python 2.5 error message is more helpful.  We
                # construct the 2.5-style message explicitly here so
                # that both Python 2.4.X and Python 2.5+ will raise
                # the same error.  This makes writing the tests eaiser
                # and makes the output easier to understand.
                raise TypeError("%r object is not iterable" %
                                type(iterable).__name__)

        length = len(iterable)
        iterator = iter(iterable)

        # insert item into repeat-dictionary
        self[key] = iterator, length, None

        return iterator, length

    def __getitem__(self, key):
        iterator, length, repeat = dict.__getitem__(self, key)
        if repeat is None:
            repeat = repeatitem(iterator, length)
            self[key] = iterator, length, repeat
        return repeat

    __getattr__ = __getitem__

    def get(self, key, default):
        if key not in self:
            return default
        return self[key]

class odict(UserDict):
    def __init__(self, dict = None):
        self._keys = []
        UserDict.__init__(self, dict)

    def __delitem__(self, key):
        UserDict.__delitem__(self, key)
        self._keys.remove(key)

    def __setitem__(self, key, item):
        """Case insensitive set item."""

        keys = tuple(key.lower() for key in self._keys)
        _key = key.lower()
        if _key in keys:
            for k in self._keys:
                if k == _key:
                    key = k
                    break
                elif k.lower() == _key:
                    self._keys.remove(k)
                    key = k
                    break

        UserDict.__setitem__(self, key, item)
        
        if key not in self._keys:
            self._keys.append(key)

    def __iter__(self):
        return iter(self._keys)
    
    def clear(self):
        UserDict.clear(self)
        self._keys = []

    def copy(self):
        dict = UserDict.copy(self)
        dict._keys = self._keys[:]
        return dict

    def items(self):
        return zip(self._keys, self.values())

    def keys(self):
        return self._keys

    def popitem(self):
        try:
            key = self._keys[-1]
        except IndexError:
            raise KeyError('dictionary is empty')

        val = self[key]
        del self[key]

        return (key, val)

    def setdefault(self, key, failobj = None):
        UserDict.setdefault(self, key, failobj)
        if key not in self._keys: self._keys.append(key)

    def update(self, dict):
        UserDict.update(self, dict)
        for key in dict.keys():
            if key not in self._keys: self._keys.append(key)

    def values(self):
        return map(self.get, self._keys)

def dotted_name(cls):
    return "%s.%s" % (cls.__module__, cls.__name__)

def get_attributes_from_namespace(element, namespace):
    if namespace is None:
        attrs = dict([
            (name, value) for (name, value) in element.attrib.items() \
            if '{' not in name])
    else:
        attrs = dict([
            (name, value) for (name, value) in element.attrib.items() \
            if name.startswith('{%s}' % namespace)])

    return attrs

def format_attribute(element, key):
    if '}' not in key:
        return key
    ns, name = key[1:].split('}')
    for prefix, namespace in element.nsmap.items():
        if ns == namespace:
            return '%s:%s' % (prefix, name)
    for prefix, namespace in config.DEFAULT_NS_MAP.items():
        if ns == namespace:
            return '%s:%s' % (prefix, name)
    raise ValueError("Can't format attribute: %s." % key)

def get_namespace(element):
    if not isinstance(element.tag, basestring):
        return
    if '}' in element.tag:
        return element.tag.split('}')[0][1:]
    return element.nsmap.get(None)

def format_kwargs(kwargs):
    items = []
    for name, value in kwargs.items():
        if isinstance(value, (int, float, basestring)):
            items.append((name, value))
        elif isinstance(value, dict):
            items.append((name, '{...} (%d)' % len(value)))
        else:
            items.append((name,
                "<%s %s at %s>" % (
                    type(value).__name__,
                    getattr(value, '__name__', "-"),
                    hex(id(value)))))

    return ["%s: %s" % item for item in items]

def raise_exc(exc):
    raise exc

def raise_template_exception(description, kwargs, exc_info):
    """Re-raise exception raised while calling ``template``, given by
    the ``exc_info`` tuple (see ``sys.exc_info``)."""

    kwargs = kwargs.copy()

    # merge dynamic scope
    kwargs.update(kwargs.pop('econtext', {}))

    # omit keyword arguments that begin with an underscore; these are
    # used internally be the template engine and should not be exposed
    map(kwargs.__delitem__, [k for k in kwargs if k.startswith('_')])

    # format keyword arguments; consecutive arguments are indented for
    # readability
    try:
        formatted_arguments = format_kwargs(kwargs)
    except:
        # the ``pprint.pformat`` method calls the representation
        # method of the arguments; this may fail and since we're
        # already in an exception handler, there's no point in
        # pursuing this further
        formatted_arguments = ()

    for index, string in enumerate(formatted_arguments[1:]):
        formatted_arguments[index+1] = " "*15 + string

    # extract line number from traceback object
    cls, exc, tb = exc_info
    try:
        lineno = tb.tb_next.tb_frame.f_lineno - 1
        filename = tb.tb_next.tb_frame.f_code.co_filename
        source = open(filename).read()
    except AttributeError:
        source = None

    # locate source code annotation (these are available from
    # the template source as comments)
    if source:
        source = source.split('\n')
        for i in reversed(range(lineno)):
            try:
                line = source[i]
            except IndexError:
                raise cls, exc, tb
            m = re_annotation.match(line)
            if m is not None:
                annotation = m.group(1)
                break
        else:
            annotation = ""

        error_msg = "Caught exception rendering template."
    else:
        annotation = ""
        error_msg = "Caught exception processing template: %s.\n" % description

    __dict__ = exc.__dict__
    __name__ = cls.__name__
    
    error_string = str(exc)

    if issubclass(cls, Exception):
        class RuntimeError(cls):
            def __init__(self, *args, **kwargs):
                pass
            
            def __str__(self):
                return "%s\n%s\n%s: %s" % (
                    error_msg, str(self.__debug_info__),
                    __name__, error_string)

        if isinstance(cls, types.TypeType):
            exc = RuntimeError.__new__(RuntimeError)
            exc.__dict__.update(__dict__)
        else:
            cls = RuntimeError

        # add/extend debug info
        try:
            debug_info = RuntimeError.__debug_info__
        except AttributeError:
            debug_info = RuntimeError.__debug_info__ = DebugInfo()
        debug_info.append(annotation, description, formatted_arguments)

    raise cls, exc, tb

class DebugInfo(object):
    def __init__(self):
        self.info = []
        
    def append(self, *args):
        self.info.append(args)

    def __str__(self):
        error_format = ' - Expression: "%s"\n' \
                       ' - Instance:   %s\n' \
                       ' - Arguments:  %s\n'

        annotation, description, formatted_arguments = self.info[-1]
        return error_format % (
            annotation, description, "\n".join(formatted_arguments))

class ValidationError(ValueError):
    def __str__(self):
        value, = self.args
        return "Insertion of %s is not allowed." % \
               repr(value)

# importing these here to avoid circular reference

import config

def xhtml_attr(name):
    return "{%s}%s" % (config.XHTML_NS, name)

def tal_attr(name):
    return "{%s}%s" % (config.TAL_NS, name)

def meta_attr(name):
    return "{%s}%s" % (config.META_NS, name)

def metal_attr(name):
    return "{%s}%s" % (config.METAL_NS, name)

def i18n_attr(name):
    return "{%s}%s" % (config.I18N_NS, name)

def py_attr(name):
    return "{%s}%s" % (config.PY_NS, name)



from chameleon.core.loader import TemplateLoader as BaseLoader
from chameleon.genshi import language
from chameleon.genshi import template


class TemplateLoader(BaseLoader):
    default_parser = language.Parser()

    formats = { "xml"  : template.GenshiTemplateFile,
                "text" : template.GenshiTextTemplateFile,
               }

    def load(self, filename, format="xml"):
        """Load and return a template file. The format parameter determines
        will parse the file. Valid options are `xml` and `text`."""
        return super(TemplateLoader, self).load(filename, self.formats[format])





import itertools

from chameleon.core import translation
from chameleon.core import config
from chameleon.core import etree
from chameleon.core import types
from chameleon.core import utils

import expressions

class ZopePageTemplateElement(translation.Element):
    """Zope Page Template element.

    Implements the ZPT subset of the attribute template language.
    """

    strip_text = False

    class node(translation.Node):
        symbols = translation.Node.symbols(
            macros=config.TRANSIENT_SYMBOL)

        content_symbol = '_content'

        ns_omit = (
            "http://xml.zope.org/namespaces/meta",
            "http://xml.zope.org/namespaces/tal",
            "http://xml.zope.org/namespaces/metal",
            "http://xml.zope.org/namespaces/i18n")

        @property
        def _interpolation_enabled(self):
            return self.element.meta_interpolation in config.TRUEVALS + ('',)

        @property
        def _interpolation_escape(self):
            return self.element.meta_interpolation_escaping in config.TRUEVALS + ('',)

        @property
        def omit(self):
            if self.element.tal_omit is not None:
                return self.element.tal_omit or True
            if self.element.meta_omit is not None:
                return self.element.meta_omit or True
            if self.element.tal_replace or self.element.meta_replace:
                return True
            if self.element.metal_use or self.element.metal_extend:
                return True

        @property
        def define(self):
            return self.element.tal_define

        @property
        def assign(self):
            content = self._content
            if content is not None:
                definition = (
                    types.declaration((self.content_symbol,)),
                    content)
                return types.definitions((definition,))

        @property
        def condition(self):
            return self.element.tal_condition

        @property
        def repeat(self):
            return self.element.tal_repeat

        @property
        def content(self):
            content = self._content
            if content is not None:
                if isinstance(content, types.escape):
                    return types.escape((types.value(self.content_symbol),))
                return types.parts((types.value(self.content_symbol),))

        @property
        def _content(self):
            return (self.element.tal_content or
                    self.element.tal_replace or
                    self.element.meta_replace)

        @property
        def skip(self):
            return (bool(self.content) or
                    bool(self.use_macro) or
                    bool(self.extend_macro) or
                    self.translate is not None)

        @property
        def dynamic_attributes(self):
            attributes = []

            tal_attributes = utils.get_attributes_from_namespace(
                self.element, config.TAL_NS)
            metal_attributes = utils.get_attributes_from_namespace(
                self.element, config.METAL_NS)
            meta_attributes = utils.get_attributes_from_namespace(
                self.element, config.META_NS)
            i18n_attributes = utils.get_attributes_from_namespace(
                self.element, config.I18N_NS)

            internal = tuple(itertools.chain(
                tal_attributes, metal_attributes, meta_attributes, i18n_attributes))

            if self._interpolation_enabled:
                attributes.extend(self.interpolated_attributes(internal))

            if self.element.tal_attributes is not None:
                attributes.extend(self.element.tal_attributes)

            if self.element.meta_attributes is not None:
                attributes.extend(self.element.meta_attributes)

            if len(attributes) > 0:
                return attributes

        @property
        def translated_attributes(self):
            return self.element.i18n_attributes

        @property
        def translate(self):
            return self.element.i18n_translate

        @property
        def translation_name(self):
            parent = self.element.getparent()
            if parent is not None and parent.node.translate is not None:
                return self.element.i18n_name

        @property
        def translation_domain(self):
            return self.element.i18n_domain

        @property
        def use_macro(self):
            return self.element.metal_use

        @property
        def define_macro(self):
            return self.element.metal_define

        @property
        def extend_macro(self):
            return self.element.metal_extend

        @property
        def define_slot(self):
            return self.element.metal_defineslot

        @property
        def fill_slot(self):
            return self.element.metal_fillslot

        @property
        def cdata(self):
            return self.element.meta_cdata

        @property
        def text(self):
            text = self.element.text
            if text is not None:
                parent = self.element.getparent()
                if parent is not None and self.element.strip_text and parent[0] is self.element:
                    text = text.lstrip('\n ')

                if not self._interpolation_enabled:
                    return (text,)

                parts = self.element.translator.split(text)
                if self.element.tag == '{http://www.w3.org/1999/xhtml}cdata':
                    return parts

                if not self._interpolation_escape:
                    return parts

                return tuple(
                    isinstance(part, types.expression) and types.escape(part) or \
                    self.element.meta_structure is True and part or \
                    utils.htmlescape(part) for part in parts)

            return ()

        @property
        def tail(self):
            tail = self.element.tail
            if tail is not None:
                if self.element.strip_text:
                    parent = self.element.getparent()
                    if parent is not None and parent.strip_text and parent[-1] is self.element:
                        tail = tail.rstrip('\n ')

                parent = self.element.getparent()
                if parent is not None and not parent.node._interpolation_enabled:
                    return (tail,)

                parts = self.element.translator.split(tail)
                if self.element.tag == '{http://www.w3.org/1999/xhtml}cdata':
                    return parts

                if not self._interpolation_escape:
                    return parts

                return tuple(
                    isinstance(part, types.expression) and types.escape(part) or \
                    self.element.meta_structure is True and part or \
                    utils.htmlescape(part) for part in parts)

            return ()

    node = property(node)

    @property
    def translator(self):
        element = self.root
        if element is None:
            raise ValueError("Default expression not set.")

        return expressions.lookup_translator(None, element.meta_translator)

    metal_define = None
    metal_use = None
    metal_extend = None
    metal_fillslot = None
    metal_defineslot = None

    i18n_name = None
    i18n_domain = None
    i18n_translate = None
    i18n_attributes = None

    tal_define = None
    tal_condition = None
    tal_replace = None
    tal_content = None
    tal_repeat = None
    tal_attributes = None

    meta_translator = etree.Annotation(
        utils.meta_attr('translator'))
    meta_interpolation = utils.attribute(
        utils.meta_attr('interpolation'), default='true', recursive=True)
    meta_interpolation_escaping = utils.attribute(
        utils.meta_attr('interpolation-escaping'), default='true', recursive=True)

class XHTMLElement(ZopePageTemplateElement):
    """XHTML namespace element."""

    tal_define = utils.attribute(
        utils.tal_attr('define'), lambda p: p.definitions)
    tal_condition = utils.attribute(
        utils.tal_attr('condition'), lambda p: p.tales)
    tal_repeat = utils.attribute(
        utils.tal_attr('repeat'), lambda p: p.definition)
    tal_attributes = utils.attribute(
        utils.tal_attr('attributes'), lambda p: p.definitions)
    tal_content = utils.attribute(
        utils.tal_attr('content'), lambda p: p.output)
    tal_replace = utils.attribute(
        utils.tal_attr('replace'), lambda p: p.output)
    tal_omit = utils.attribute(
        utils.tal_attr('omit-tag'), lambda p: p.tales)
    tal_default_expression = utils.attribute(
        utils.tal_attr('default-expression'), encoding='ascii')
    metal_define = utils.attribute(
        utils.metal_attr('define-macro'))
    metal_use = utils.attribute(
        utils.metal_attr('use-macro'), lambda p: p.tales)
    metal_extend = utils.attribute(
        utils.metal_attr('extend-macro'), lambda p: p.tales)
    metal_fillslot = utils.attribute(
        utils.metal_attr('fill-slot'))
    metal_defineslot = utils.attribute(
        utils.metal_attr('define-slot'))
    i18n_translate = utils.attribute(
        utils.i18n_attr('translate'))
    i18n_attributes = utils.attribute(
        utils.i18n_attr('attributes'), lambda p: p.mapping)
    i18n_domain = utils.attribute(
        utils.i18n_attr('domain'))
    i18n_name = utils.attribute(
        utils.i18n_attr('name'))
    
class MetaElement(XHTMLElement, translation.MetaElement):
    pass

class TALElement(XHTMLElement):
    """TAL namespace element."""

    strip_text = True
    
    tal_define = utils.attribute(
        ("define", utils.tal_attr("define")), lambda p: p.definitions)
    tal_condition = utils.attribute(
        ("condition", utils.tal_attr("condition")), lambda p: p.tales)
    tal_replace = utils.attribute(
        ("replace", utils.tal_attr("replace")), lambda p: p.output)
    tal_repeat = utils.attribute(
        ("repeat", utils.tal_attr("repeat")), lambda p: p.definition)
    tal_attributes = utils.attribute(
        ("attributes", utils.tal_attr("attributes")),
        lambda p: utils.raise_exc(
            TypeError("Dynamic attributes not allowed on "
                      "elements with TAL-namespace.")))
    tal_content = utils.attribute(
        ("content", utils.tal_attr("content")), lambda p: p.output)
    tal_omit = utils.attribute(
        ("omit-tag", utils.tal_attr("omit-tag")), lambda p: p.tales, u"")    

class METALElement(XHTMLElement):
    """METAL namespace element."""

    strip_text = True
    tal_omit = True
    
    metal_define = utils.attribute(
        ("define-macro", utils.metal_attr("define-macro")))
    metal_use = utils.attribute(
        ('use-macro', utils.metal_attr('use-macro')), lambda p: p.tales)
    metal_extend = utils.attribute(
        ('extend-macro', utils.metal_attr('extend-macro')), lambda p: p.tales)
    metal_fillslot = utils.attribute(
        ('fill-slot', utils.metal_attr('fill-slot')))
    metal_defineslot = utils.attribute(
        ('define-slot', utils.metal_attr('define-slot')))

class TextElement(XHTMLElement):
    meta_interpolation_escaping = False

class Parser(etree.Parser):
    """Zope Page Template parser."""
    
    element_mapping = {
        config.XHTML_NS: {None: XHTMLElement},
        config.META_NS: {None: MetaElement},
        config.TAL_NS: {None: TALElement},
        config.METAL_NS: {None: METALElement}}

    fallback = XHTMLElement
    
    default_expression = 'python'

    def __init__(self, default_expression=None):
        if default_expression is not None:
            self.default_expression = default_expression

    def parse(self, body):
        tree = super(Parser, self).parse(body)
        tree.getroot().meta_translator = self.default_expression
        return tree

    def validate(self, element):
        kls = element.__class__
        if kls.metal_use.get(element) and kls.metal_define.get(element):
            raise SyntaxError(
                "Cannot use `define-macro` with `use-macro`; use "
                "`extend-macro` instead.")


class TextParser(Parser):
    """Parser for text templates."""

    element_mapping = {
        config.XHTML_NS: {None: TextElement},
    }

